"""Logging setup for autopypath package."""

import logging

__all__ = []



logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_log = logging.getLogger('autopypath')
"""Logging instance for the autopypath package."""



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# pylint: disable=invalid-name, missing-module-docstring
from __future__ import annotations

import sys
from pathlib import Path

sys.path.insert(0, str(Path('..', 'src').resolve()))
sys.path.insert(0, str(Path('.').resolve()))

from autopypath import _meta

# -- Project information -----------------------------------------------------

project: str = _meta.__project__
copyright: str = _meta.__copyright__  # noqa: A001
author: str = _meta.__author__

# The short X.Y version
version: str = _meta.__version__
# The full version, including alpha/beta/rc tags
release: str = _meta.__release__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # 'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.apidoc',
    # 'sphinx.ext.autodoc',
    # 'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    # 'sphinx.ext.napoleon',
    'sphinx.ext.githubpages',
    'sphinx.ext.viewcode',
    'sphinx_copybutton',
    'sphinx_design',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
    'pytest': ('https://docs.pytest.org/en/latest/', None),
    'tox': ('https://tox.readthedocs.io/en/latest/', None),
}
intersphinx_disabled_domains = ['std']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'tests',
    'conf.pycoverage',
    '.venv',
    '.venv-3.10',
    '.venv-3.11',
    '.venv-3.12',
    '.venv-3.13',
    '.venv-3.14',
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'github-dark'


# -- Options for HTML output -------------------------------------------------

# Maximum line length for signatures in the HTML docs.
maximum_signature_line_length: int = 50

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'sphinxawesome_theme'
html_theme: str = 'furo'
# html_logo = '_static/images/simplebench-logo.svg'


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom CSS files.
html_css_files = ['custom.css']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# Icon to use for the permalink to the current page.
html_permalinks_icon = '🔗'

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'simplebenchdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {  # type: ignore [var-annotated]
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, 'simplebench.tex', 'simplebench Documentation', 'Jerilyn Franz', 'manual')]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual metric).
man_pages = [(master_doc, 'simplebench', 'simplebench Documentation', [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'simplebench',
        'simplebench Documentation',
        author,
        'simplebench',
        'One line description of project.',
        'Miscellaneous',
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
# epub_title = simplebench

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

epub_show_urls = 'footnote'

# -- Extension configuration -------------------------------------------------

# By default, highlight as Python 3.
highlight_language = 'python3'

# This pattern will automatically exclude prompts from shell blocks
copybutton_prompt_text = r'>>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: '
copybutton_prompt_is_regexp = True


# -- Options for autodoc extension -------------------------------------------

# This value determines whether module names are prepended to all
# documented members. If true, the module name is prepended.
# autodoc_canonical_imports = True

# This value controls how to represent the signature of a class.
# 'class' shows the class name, 'init' shows the __init__ signature.
autodoc_class_signature = 'mixed'

# This value selects what content will be inserted into the main body of an
# autoclass directive. The possible values are:
# "class", "init", "both"
autoclass_content = 'both'


# -- Options for doctest -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/doctest.html#configuration

# This code block will be executed before each doctest.
# We use it to inject our custom assertion helper into the test's global namespace.
doctest_global_setup = """
from _helpers.doctest_utils import run_script_and_get_raw_output, assert_benchmark_output
"""



"""Enables debug logging for autopypath when imported.

**Automatic Debug Mode**
------------------------
By importing the :mod:`autopypath.debug` submodule instead of :mod:`autopypath`,
detailed debug logging is enabled to trace how the project root is determined,
which paths are added to `sys.path`, and any issues encountered along the way.

Additionally, warnings logged by the configuration processing are upgraded to logged errors
and exceptions are raised for them. This does not affect warnings originating from other modules
that autopypath depends on.

It changes the global logging level for the :mod:`autopypath._log` logger to ``DEBUG``.
Normally this should have no side effects because autopypath can only be imported once per process
and does not publically expose any logging configuration APIs.

Otherwise, it is functionally equivalent to importing :mod:`autopypath`.

This is useful for troubleshooting and understanding the internal workings of autopypath.

.. code-block:: python

    import autopypath.debug
    # sys.path was adjusted automatically with debug logging and exceptions enabled

"""

import logging
from pathlib import Path
from typing import Optional

from ._config_py_path import _ConfigPyPath
from ._context import _context_frameinfo
from ._exceptions import AutopypathError
from ._log import _log

__all__ = ['AutopypathError']

_log.setLevel(logging.DEBUG)


_context_file: Optional[Path] = None
"""This is the file path of the script that imported this module, if available."""
_context_name: Optional[str] = None
"""This is the __name__ of the script that imported this module, if available."""
_path_adjusted: bool = False
"""Indicates whether autopypath adjusted sys.path automatically upon import."""

_context_info: Optional[tuple[Path, str]] = _context_frameinfo()
if _context_info is not None:
    _context_file, _context_name = _context_info
    if _context_name != '__main__':
        _log.debug(
            'autopypath imported from non-__main__ context (%s); no sys.path changes will be applied.',
            _context_name,
        )
        _path_adjusted = False
    else:
        _ConfigPyPath(context_file=_context_file, strict=True)
        _log.debug('sys.path adjusted automatically for %s', _context_file)
        _path_adjusted = True
else:
    _log.error('could not determine context file; no sys.path changes will be applied.')
    raise AutopypathError('could not determine context file for autopypath; sys.path not adjusted.')



"""TestSpec testing framework - Pytest shim for test actions.

This module provides a PytestAction class that serves as a bridge between
the TestAction class and the pytest testing framework. It allows for the
declarative specification of tests while assigning unique identifiers to
each test case for Pytest.
"""

from collections.abc import Callable
from typing import Any, NoReturn, Optional, Union

from .assertions import Assert
from .base import TestSpec
from .constants import NO_EXPECTED_VALUE
from .helpers import no_assigned_action
from .idspec import idspec
from .test_action import TestAction


class PytestAction(TestSpec):
    """A generic unit test specification class for pytest actions.

    It allows tests to be specified declaratively while providing a large amount
    of flexibility. This is a thin wrapper around :class:`TestAction` that adds pytest-specific
    functionality, such as assigning unique ids to each test case.

    :param str ident: Id for the test.
    :param str name: Identifying name for the test.
    :param Optional[Callable[..., Any]] action: A reference to a callable function or method to be invoked for the test.
                    If no action is assigned, the special function `no_assigned_action` is used which
                    raises NotImplementedError when called.
                    Defaults to no_assigned_action.
    :param Optional[list[Any]] args: Sequence of positional arguments to be passed to the `action` function or method.
                    Defaults to [].
    :param Optional[dict[str, Any]] kwargs: Dictionary containing keyword arguments to be passed to the `action`
                    function or method. Defaults to {}.
    :param Optional[Assert] assertion: The assertion operator to use when comparing the expected and found values.
                        Defaults to Assert.EQUAL.
    :param Any expected: Expected value (if any) for the `action` function or method.
                        This is used with the `assertion` operator to validate the return value of the
                        function or method.
                        If there is no expected value, the special class NoExpectedValue is used to flag it.
                        This is used so that the specific return value of None can be distinguished from no
                        particular value or any value at all is expected to be returned from the function or method.
                        Defaults to NO_EXPECTED_VALUE.
    :param Optional[Any] obj: Optional object to be validated. Defaults to None.
    :param Optional[Callable[[Any], bool]] validate_obj: Function to validate the optional object. Defaults to None.
    :param Optional[Callable[[Any], bool]] validate_result: Function to validate the result of the action.
                        Defaults to None.
    :param Optional[str] validate_attr: Validate a property of the result instead of the result itself.
                        Defaults to None.
    :param Optional[type[BaseException]] exception: Expected exception type (if any) to be raised by the action.
                        Defaults to None.
    :param Optional[str] exception_tag: Expected tag (if any) to be found in the exception message. Defaults to None.
    :param Optional[Callable[[str], NoReturn]] on_fail: Function to call on test failure. Defaults to _fail
                        method which raises AssertionError.
    :param Optional[Any] extra: Extra data for use by test frameworks. It is not used by the TestAction class itself.
                    Defaults to None.
    """

    __test__ = False  # Prevent pytest from collecting this class as a test case

    def __new__(
        cls,
        ident: str,
        *,
        name: str = '',
        action: Callable[..., Any] = no_assigned_action,
        args: Optional[list[Any]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        assertion: Assert = Assert.EQUAL,
        expected: Any = NO_EXPECTED_VALUE,
        obj: Optional[Any] = None,
        validate_obj: Optional[Callable[[Any], bool]] = None,
        validate_result: Optional[Callable[[Any], bool]] = None,
        validate_attr: Optional[str] = None,
        exception: Optional[type[BaseException]] = None,
        exception_tag: Optional[str] = None,
        display_on_fail: Union[str, Callable[[], str]] = '',
        on_fail: Optional[Callable[[str], NoReturn]] = None,
        extra: Any = None,
    ) -> Any:
        """Run the test action using pytest.


        It allow tests to be specified declaratively while providing a large amount
        of flexibility.
        :param str ident: Id for the test.
        :param str name: Identifying name for the test.
        :param Optional[Callable[..., Any]] action: A reference to a callable function or method to be invoked
                        for the test.
                        If no action is assigned, the special function `no_assigned_action` is used which
                        raises NotImplementedError when called.
                        Defaults to no_assigned_action.
        :param Optional[list[Any]] args: Sequence of positional arguments to be passed to the `action` function
                        or method. Defaults to [].
        :param Optional[dict[str, Any]] kwargs: Dictionary containing keyword arguments to be passed to the `action`
                        function or method. Defaults to {}.
        :param Optional[Assert] assertion: The assertion operator to use when comparing the expected and found values.
                          Defaults to Assert.EQUAL.
        :param Any expected: Expected value (if any) for the `action` function or method.
                         This is used with the `assertion` operator to validate the return value of the
                         function or method.
                         If there is no expected value, the special class NoExpectedValue is used to flag it.
                         This is used so that the specific return value of None can be distinguished from no
                         particular value or any value at all is expected to be returned from the function or method.
                         Defaults to NO_EXPECTED_VALUE.
        :param Optional[Any] obj: Optional object to be validated. Defaults to None.
        :param Optional[Callable[[Any], bool]] validate_obj: Function to validate the optional object.
                        Defaults to None.
        :param Optional[Callable[[Any], bool]] validate_result: Function to validate the result of the action.
                        Defaults to None.
        :param Optional[str] validate_attr: Validate a property of the result instead of the result itself.
                        Defaults to None.
        :param Optional[type[BaseException]] exception: Expected exception type (if any) to be raised by the action.
                        Defaults to None.
        :param Optional[str] exception_tag: Expected tag (if any) to be found in the exception message.
                        Defaults to None.
        :param Optional[Callable[[str], NoReturn]] on_fail: Function to call on test failure.
                        Defaults to _fail method which raises AssertionError.
        :param Optional[Any] extra: Extra data for use by test frameworks. It is not used by the TestAction class
                        itself. Defaults to None.
        """
        return idspec(
            ident,
            TestAction(
                name=name,
                action=action,
                args=args,
                kwargs=kwargs,
                assertion=assertion,
                expected=expected,
                obj=obj,
                validate_obj=validate_obj,
                validate_result=validate_result,
                validate_attr=validate_attr,
                exception=exception,
                exception_tag=exception_tag,
                display_on_fail=display_on_fail,
                on_fail=on_fail,
                extra=extra,
            ),
        )

    def run(self) -> Any:
        """Run the test action using pytest.

        :return: The result of the test action.
        :rtype: Any
        """
        raise NotImplementedError(
            'PytestAction instances are not meant to be run directly.')



"""Core configuration functionality for AutoPyPath."""
# ruff: noqa F401

from ._config import (
    _AutopypathConfig,
    _Config,
    _DefaultConfig,
    _ManualConfig,
    _PyProjectConfig,
    _TomlConfig,
)
from ._config_py_path import _ConfigPyPath

# No '*' exports
__all__ = []



"""Decorators for enums."""

import ast
import inspect
import textwrap
from enum import Enum
from functools import partial
from operator import is_
from typing import TypeVar, Union

E = TypeVar('E', bound=Enum)


# Decorator to attach docstrings to enum members
# See: https://stackoverflow.com/questions/19330460/how-do-i-put-docstrings-on-enums
def enum_docstrings(enum: type[E]) -> type[E]:
    """Attach docstrings to enum members.

    Docstrings are string literals that appear directly below the enum member
    assignment expression within triple-quotes.

    This decorator parses the source code of the enum class to find
    docstrings for each member and attaches them to the respective enum members.

    This allows for more detailed documentation of enum members and in tools
    that can extract and display these docstrings.

    This code is adapted from:
    https://stackoverflow.com/questions/19330460/how-do-i-put-docstrings-on-enums

    .. code-block:: python
      :caption: Example usage of enum_docstrings decorator
      :linenos:

      @enum_docstrings
      class SomeEnum(Enum):
          '''Docstring for the SomeEnum enum'''

          foo_member = 'foo_value'
          '''Docstring for the foo_member enum member'''


      SomeEnum.foo_member.__doc__  # 'Docstring for the foo_member enum member'

    :param enum: The enum class to process.
    :return: The same enum class with member docstrings attached.
    """
    if not issubclass(enum, Enum):
        raise TypeError('enum_docstrings can only be applied to Enum subclasses.')

    try:
        source = inspect.getsource(enum)
        source = textwrap.dedent(source)
        mod = ast.parse(source)
    except OSError:  # Fallback case where source code is not available
        return enum

    # with the source parsed, find the class definition
    # if there is no class definition, return the enum unmodified
    if not (mod.body and isinstance(class_def := mod.body[0], ast.ClassDef)):
        return enum

    # An enum member docstring is unassigned if it is the exact same object
    # as enum.__doc__.
    unassigned = partial(is_, enum.__doc__)
    names = enum.__members__.keys()
    member: Union[E, None] = None
    for node in class_def.body:
        if (
            isinstance(node, ast.Assign)
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and (name := node.targets[0].id) in names
        ):
            # Enum member assignment, look for a docstring next
            member = enum[name]
            continue

        elif (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and isinstance(node.value.value, str)
            and member
            and unassigned(member.__doc__)
        ):
            # docstring immediately following a member assignment
            member.__doc__ = node.value.value

        else:
            pass

        member = None

    return enum



"""Configuration modules for autopypath.

This module contains various configuration classes for autopypath,
each designed to load configuration settings from different sources.

The available configuration classes include:

- :class:`Config`: Base configuration class.
- :class:`TomlConfig`: Base class for TOML-based configurations.
- :class:`AutopypathConfig`: Configuration from autopypath.toml files.
- :class:`DefaultConfig`: Default configuration.
- :class:`ManualConfig`: Configuration from manual paths.
- :class:`PyProjectConfig`: Configuration from pyproject.toml files.
"""
# ruff: noqa F401

from ._autopypath import _AutopypathConfig
from ._config import _Config
from ._default import _DefaultConfig
from ._manual import _ManualConfig
from ._pyproject import _PyProjectConfig
from ._toml import _TomlConfig

__all__ = []



"""TestSpec framework - TestGet class."""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import Any, Callable, NoReturn, Optional, Union

from .assertions import Assert, validate_assertion
from .base import TestSpec
from .constants import NO_EXPECTED_VALUE, NO_OBJ_ASSIGNED
from .helpers import _process_exception


@dataclass
class TestGet(TestSpec):
    """A class for testing getting attributes.

    This class allows for testing getting of attributes on an object,
    including validation of the get value, expected exceptions, and custom validation functions.

    :param name: Identifying name for the test.
    :type name: str
    :param obj: The object whose attribute is to be tested. If not provided, the special sentinel value
                NO_OBJ_ASSIGNED is used. The object must not be None and must be an instance of object.
                It must be provided to run the test.
                Defaults to NO_OBJ_ASSIGNED.
    :type obj: Optional[object], optional
    :param attribute: The name of the attribute to be tested by setting.
    :type attribute: str
    :param expected: Expected value of attribute after setting the attribute. If a exception is set,
                     the expected value is ignored.
    :type expected: Optional[Any], optional
    :param assertion: The assertion operator to use when comparing the expected and found values.
                      Defaults to Assert.EQUAL.
    :type assertion: Assert, optional
    :param exception: Expected exception type (if any) to be raised by setting the attribute.
                      Defaults to None.
    :type exception: Optional[type[Exception]], optional
    :param exception_tag: Expected tag (if any) to be found in the exception message.
                          Defaults to None.
    :type exception_tag: Optional[str | Enum], optional
    :param validate: Function to validate obj after getting attribute. It should return True if the object state is
                     valid. This is distinguished from the expected value check in that it can perform more complex
                     validation of the entire object state rather than just checking the value of a single attribute.
                     It is passed two arguments, the TestGet instance and the object being validated.
                     Defaults to None.
    :type validate: Optional[Callable[[TestGet, Any], bool]], optional
    :param display_on_fail: Message to display on test failure.
                            This can be a static string or a callable that returns a string.
                            Defaults to "".
    :type display_on_fail: str | Callable[[], str], optional
    :param Callable[[str], NoReturn], optional on_fail: Function to call on test failure.
        Defaults to _fail method which raises AssertionError.
    :param extra: Extra fields for use by test frameworks. It is not used by the TestGet class itself.
                  Defaults to None.
    :type extra: Any, optional
    """
    __test__ = False  # Prevent pytest from trying to collect this class as a test case

    name: str
    """Identifying name for the test."""
    attribute: str
    """The name of the attribute to be tested by setting."""
    obj: Optional[object] = NO_OBJ_ASSIGNED
    """The object whose attribute is to be tested. It cannot be None, and must be an instance of object.
    If not provided during construction, the special sentinel value NO_OBJ_ASSIGNED is used. This must be
    replaced with a valid object before running the test."""
    assertion: Assert = Assert.EQUAL
    """The assertion operator to use when comparing the expected and found values. (default is Assert.EQUAL)"""
    expected: Optional[Any] = NO_EXPECTED_VALUE
    """Expected value of attribute after setting the attribute. If a get_exception or
    set_exception is set, the expected value is ignored. If there is no expected exception
    and no expected value, use the special sentinel value NO_EXPECTED_VALUE to skip the
    get step validation for the set value. Omitting this field is equivalent to setting it to NO_EXPECTED_VALUE."""
    exception: Optional[type[BaseException]] = None
    """Expected exception type (if any) to be raised by getting the attribute."""
    exception_tag: Optional[Union[str, Enum]] = None
    """Expected tag (if any) to be found in an exception message raised by getting the attribute."""
    validate: Optional[Callable[[TestGet, Any], Union[None, NoReturn]]] = None
    """Function to validate obj state after setting the attribute. It should raise an exception
    if the object state is unexpected.

    This is distinguished from the expected value check in that it can perform more complex validation
    of the entire object state rather than just checking the value of a single attribute.

    It is passed two arguments, the TestSetGet instance and the object being validated.

    The validation function should call the `on_fail` method to raise an exception if the object is not
    in a valid state. None should be returned if the object is valid.
    """
    display_on_fail: Union[str, Callable[[], str]] = ""
    """String or function to display additional information on test failure."""
    on_fail: Optional[Callable[[str], NoReturn]] = None
    """Function to call on test failure. The function should raise an exception (default is pytest.fail)."""
    extra: Any = None
    """Extra data for use by test frameworks. It is not used by the TestGet class itself. Default is None."""

    _creation_traceback: Optional[TracebackType] = None
    """The traceback at the point where the TestAction was created."""

    def __post_init__(self) -> None:
        """Post-initialization validation checks."""
        if not isinstance(self.name, str):
            raise TypeError("name must be a str")
        if not isinstance(self.attribute, str):
            raise TypeError("attribute must be a str")
        if self.obj is None:
            raise TypeError("obj cannot be None")
        if not isinstance(self.obj, object):
            raise TypeError("obj must be an object")
        if self.assertion not in Assert:
            raise ValueError(f"assertion must be one of {list(Assert)}")
        if self.attribute == "":
            raise ValueError("attribute cannot be an empty string")
        if self.validate is not None and not callable(self.validate):
            raise TypeError("validate must be callable if provided")
        if self.exception is not None and not issubclass(self.exception, Exception):
            raise TypeError("exception must be an Exception type if provided")
        if self.exception_tag is not None and not isinstance(self.exception_tag, (str, Enum)):
            raise TypeError("set_exception_tag must be a str or Enum if provided")
        if not callable(self.on_fail):
            raise TypeError("on_fail must be callable")
        if self.display_on_fail and not (isinstance(self.display_on_fail, str) or callable(self.display_on_fail)):
            raise TypeError("display_on_fail must be a str or callable if provided")
        super().__post_init__()

    def run(self) -> None:
        """Execute the attribute get test."""
        # disabled because we are using the __getattribute__ dunder method directly
        # for testing purposes because there is no other way to do get/set testing for attributes generically.
        # pylint: disable=unnecessary-dunder-call

        # hide traceback for this function in pytest output. Disabled for pylint because it is a pytest
        # feature that is just not understood by pylint.
        __tracebackhide__ = True  # pylint: disable=unused-variable

        test_description: str = f"{self.name}"

        # Errors found during the test
        errors: list[str] = []

        obj = self.obj
        expected = self.expected
        validate = self.validate
        if obj is NO_OBJ_ASSIGNED:
            if self.on_fail:
                self.on_fail(f"{self.name}: obj for test is not assigned")
            else:
                self._fail(f"{self.name}: obj for test is not assigned")
            raise RuntimeError("unreachable code after on_fail call")  # pylint: disable=raise-missing-from

        if self.exception is None and expected is NO_EXPECTED_VALUE and validate is None:
            raise ValueError("No validation is possible: exception, expected, and "
                             "validate are all None/NO_EXPECTED_VALUE or not set")

        attribute_to_get = self.attribute
        try:
            if validate is not None:
                validate(self, obj)  # Exception should be raised by validate if unexpected obj state

            if expected is NO_EXPECTED_VALUE:
                return

            found: Any = obj.__getattribute__(attribute_to_get)
            if expected is not NO_EXPECTED_VALUE:
                assertion_result = validate_assertion(self.assertion, expected, found)
                if assertion_result:
                    errors.append(assertion_result)
                    if callable(self.display_on_fail):
                        errors.append(self.display_on_fail())
                    elif isinstance(self.display_on_fail, str):
                        errors.append(self.display_on_fail)

        except BaseException as err:  # pylint: disable=broad-exception-caught
            new_errors = _process_exception(
                err=err,
                exception=self.exception,
                exception_tag=self.exception_tag,
                label="validating attribute value")
            errors.extend(new_errors)

        # Report any errors found during the get portion of the test
        if errors:
            if self.on_fail:
                self.on_fail(test_description + ": " + "\n".join(errors))
            else:
                self._fail(test_description + ": " + "\n".join(errors))



"""Marker types for repository markers."""

from enum import Enum
from types import MappingProxyType
from typing import Union

from ._doc_utils import enum_docstrings
from ._typing import Final, Literal, TypeAlias, TypeGuard

__all__ = ['_MarkerType']


@enum_docstrings
class _MarkerType(str, Enum):
    """Types of repository markers used to identify the repository root.

    - FILE: A file that must exist in the repository root.
    - DIR: A directory that must exist in the repository root.

    Example
    -------
    .. code-block:: python
        from autopypath.marker_type import MarkerType

        marker_type = MarkerType.FILE

    """

    FILE = 'file'
    """A file that must exist in the repository root."""
    DIR = 'dir'
    """A directory that must exist in the repository root."""


MarkerTypeLiteral: TypeAlias = Literal['file', 'dir']
"""Literal type for MarkerType values."""


MARKER_TYPE_MAP: Final[MappingProxyType[MarkerTypeLiteral, _MarkerType]] = MappingProxyType(
    {marker.value: marker for marker in _MarkerType}
)
"""Mapping from literal strings to MarkerType enum members.

Example
-------

.. code-block:: python

    from autopypath.marker_type import MARKER_TYPE_MAP, MarkerType

    marker_type = MARKER_TYPE_MAP['file']
    assert marker_type == MarkerType.FILE
"""


def is_marker_type_literal(value: str) -> TypeGuard[MarkerTypeLiteral]:
    """Checks if the given string is a valid MarkerType literal.

    Example
    -------

    .. code-block:: python
        from autopypath.marker_type import is_marker_type_literal

        assert is_marker_type_literal('file') is True
        assert is_marker_type_literal('dir') is True
        assert is_marker_type_literal('invalid') is False

    :param str value: The string to check.
    :return bool: ``True`` if the string is a valid ``MarkerType`` literal, ``False`` otherwise.
    """
    return value in MARKER_TYPE_MAP


def resolve_marker_type_literal(value: str) -> Union[_MarkerType, None]:
    """Resolves a string literal to its corresponding MarkerType enum member
    or returns ``None`` if the literal is invalid.

    Example
    -------

    .. code-block:: python
        from autopypath.path_resolution_order import resolve_marker_type_literal, PathResolution

        assert resolve_marker_type_literal('file') == MarkerType.FILE
        assert resolve_marker_type_literal('dir') == MarkerType.DIR
        assert resolve_marker_type_literal('invalid') is None
    :param str value: The string literal to resolve.
    :return MarkerType | None: The corresponding MarkerType enum member,
                               or ``None`` if the literal is invalid.
    """
    if is_marker_type_literal(value):
        return MARKER_TYPE_MAP.get(value)
    return None



"""Tests for autopypath._validate module."""

import logging
import os
from pathlib import Path
from types import MappingProxyType

import pytest
from testspec import Assert, PytestAction, TestSpec

from autopypath import _validate
from autopypath._exceptions import AutopypathError
from autopypath._load_strategy import _LoadStrategy
from autopypath._marker_type import _MarkerType
from autopypath._path_resolution import _PathResolution

# fmt: off

@pytest.mark.parametrize(
    'testspec', [
        PytestAction('CONTEXT_001',
            name='valid context file',
            action=_validate.context_file, args=[__file__],
            exception=AutopypathError),
        PytestAction('CONTEXT_002',
            name='non-existent context file',
            action=_validate.context_file, args=['non_existent_file.cfg'],
            exception=AutopypathError),
        PytestAction('CONTEXT_003',
            name='context file as Path',
            action=_validate.context_file, args=[Path(__file__)],
            expected=Path(__file__).resolve()),
        PytestAction('CONTEXT_004',
            name='Context file is a directory',
            action=_validate.context_file, args=[Path('.')],
            exception=AutopypathError),
        PytestAction('CONTEXT_005',
            name='context file as string relative to cwd',
            action=_validate.context_file, args=['data'],
            exception=AutopypathError),  # Relative to cwd
        PytestAction('CONTEXT_006',
            name='context file as None',
            action=_validate.context_file, args=[None],
            exception=AutopypathError),
    ])
def test_context_file(testspec: TestSpec) -> None:
    """Test Config with context_file."""
    testspec.run()


@pytest.mark.parametrize(
    'testspec', [
        PytestAction('ROOT_REPO_001',
            name='invalid root_repo',
            action=_validate.root_repo_path, args=['not-my-repo-does-not-exist'],
            exception=AutopypathError),
        PytestAction('ROOT_REPO_002',
            name='valid root_repo',
            action=_validate.root_repo_path, args=['.'],
            expected=Path('.')),
        PytestAction('ROOT_REPO_003',
            name='None root_repo',
            action=_validate.root_repo_path, args=[None],
            exception=AutopypathError),
        PytestAction('ROOT_REPO_004',
            name='root_repo as Path',
            action=_validate.root_repo_path, args=[Path('.')],
            expected=Path('.')),
        PytestAction('ROOT_REPO_005',
            name='root_repo as empty string',
            action=_validate.root_repo_path, args=[''],
            exception=AutopypathError),
        PytestAction('ROOT_REPO_006',
            name='Non-existent root_repo with multi-level path',
            action=_validate.root_repo_path, args=[Path('data','more')],
            exception=AutopypathError),
        PytestAction('ROOT_REPO_007',
            name='root_repo two levels up from current directory',
            action=_validate.root_repo_path, args=[Path('..','..')],
            expected=Path('..','..')),
        PytestAction('ROOT_REPO_008',
            name='root_repo pointing to a file instead of a directory',
            action=_validate.root_repo_path, args=[Path('data','not_a_directory')],
            exception=AutopypathError),
        PytestAction('ROOT_REPO_009',
            name='root_repo pointing to test data directory',
            action=_validate.root_repo_path, args=[Path(__file__).parent / Path('data')],
            expected=(Path(__file__).parent / Path('data')).resolve()),
        PytestAction('ROOT_REPO_010',
            name='root_repo with path include null byte',
            action=_validate.root_repo_path, args=[Path('data', 'invalid\0path')],
            exception=AutopypathError),
    ])
def test_root_repo_path(testspec: TestSpec) -> None:
    """Test Config with root_repo_path."""
    testspec.run()


@pytest.mark.parametrize(
    'testspec', [
        PytestAction('MARKER_001',
            name='valid repo_markers',
            action=_validate.repo_markers,
            args=[{'.git': _MarkerType.DIR, 'pyproject.toml': _MarkerType.FILE}],
            expected={'.git': _MarkerType.DIR, 'pyproject.toml': _MarkerType.FILE}),
        PytestAction('MARKER_002',
            name='empty repo_markers',
            action=_validate.repo_markers, args=[{}],
            expected=None),
        PytestAction('MARKER_003',
            name='single repo_marker',
            action=_validate.repo_markers, args=[{'setup.py': _MarkerType.FILE}],
            expected={'setup.py': _MarkerType.FILE}),
        PytestAction('MARKER_004',
            name='string repo_markers',
            action=_validate.repo_markers, args=[{'.git': 'dir', 'setup.py': 'file'}],
            expected={'.git': _MarkerType.DIR, 'setup.py': _MarkerType.FILE}),
        PytestAction('MARKER_005',
            name='mixed type repo_markers',
            action=_validate.repo_markers, args=[{'.git': _MarkerType.DIR, 'setup.py': 'file'}],
            expected={'.git': _MarkerType.DIR, 'setup.py': _MarkerType.FILE}),
        PytestAction('MARKER_006',
            name='invalid repo_marker',
            action=_validate.repo_markers, args=[{'.git': 'DIR'}],
            exception=AutopypathError),
        PytestAction('MARKER_007',
            name='invalid repo_marker type',
            action=_validate.repo_markers, args=[{'.git': 123}],
            exception=AutopypathError),
        PytestAction('MARKER_008',
            name='None repo_markers',
            action=_validate.repo_markers, args=[None],
            expected=None),
        PytestAction('MARKER_009',
            name='no repo_markers',
            action=_validate.repo_markers, kwargs={'value': {}},
            expected=None),
        PytestAction('MARKER_010',
            name='whitespace repo_marker key',
            action=_validate.repo_markers, args=[{'   ': _MarkerType.DIR}],
            exception=AutopypathError),
        PytestAction('MARKER_011',
            name='non-string repo_marker key',
            action=_validate.repo_markers, args=[{123: _MarkerType.DIR}],
            exception=AutopypathError),
        PytestAction('MARKER_012',
            name='non-mapping repo_markers',
            action=_validate.repo_markers, args=[[('setup.py', _MarkerType.FILE)]],
            exception=AutopypathError),
        PytestAction('MARKER_013',
            name='repo_marker key exceeding max length',
            action=_validate.repo_markers, args=[{'a' * 65: _MarkerType.DIR}],
            exception=AutopypathError),
        PytestAction('MARKER_014',
            name='repo_marker key containing path separator',
            action=_validate.repo_markers, args=[{'inva/lid': _MarkerType.DIR}],
            exception=AutopypathError),
        PytestAction('MARKER_015',
            name='repo_marker key being Windows reserved name',
            action=_validate.repo_markers, args=[{'CON': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_016',
            name='repo_marker key containing forbidden characters',
            action=_validate.repo_markers, args=[{'inva<lid': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_017',
            name='repo_marker key being empty string',
            action=_validate.repo_markers, args=[{'': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_018',
            name='repo_marker key having leading whitespace',
            action=_validate.repo_markers, args=[{'  setup.py': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_019',
            name='repo_marker key having trailing whitespace',
            action=_validate.repo_markers, args=[{'setup.py  ': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_020',
            name='repo_marker key containing null character',
            action=_validate.repo_markers, args=[{'setup\0.py': _MarkerType.FILE}],
            exception=AutopypathError),
        PytestAction('MARKER_021',
            name='Validate repo_marker attr returns MappingProxyType',
            action=_validate.repo_markers, args=[{'setup.py': _MarkerType.FILE}],
            expected=MappingProxyType,
            assertion=Assert.ISINSTANCE),
    ]
)
def test_repo_markers(testspec: TestSpec) -> None:
    """Test Config with repo_markers."""
    testspec.run()

@pytest.mark.parametrize('testspec', [
    PytestAction('PATHS_001',
        name='valid list of paths',
        action=_validate.paths, args=[['src', 'lib', 'utils']],
        expected=(Path('src'), Path('lib'), Path('utils'))),
    PytestAction('PATHS_002',
        name='empty paths',
        action=_validate.paths, args=[[]],
        expected=None),
    PytestAction('PATHS_003',
        name='single path',
        action=_validate.paths, args=[['src']],
        expected=(Path('src'),)),
    PytestAction('PATHS_004',
        name='None paths',
        action=_validate.paths, kwargs={'value': None},
        expected=None),
    PytestAction('PATHS_005',
        name='no paths',
        action=_validate.paths, kwargs={'value': []},
        expected=None),
    PytestAction('PATHS_006',
        name='invalid path type',
        action=_validate.paths, args=[['src', 123, 'utils']],
        exception=AutopypathError),
    PytestAction('PATHS_007',
        name='Path objects',
        action=_validate.paths, args=[[Path('src'), Path('lib')]],
        expected=(Path('src'), Path('lib'))),
    PytestAction('PATHS_008',
        name='path having leading whitespace',
        action=_validate.paths, args=[['  src', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_009',
        name='path having trailing whitespace',
        action=_validate.paths, args=[['src  ', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_010',
        name='path being only whitespace',
        action=_validate.paths, args=[['   ', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_011',
        name='path being only forward slashes',
        action=_validate.paths, args=[['///', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_012',
        name='path being only backslashes',
        action=_validate.paths, args=[['\\\\\\', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_013',
        name='path having segment with leading whitespace',
        action=_validate.paths, args=[['src/  utils', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_014',
        name='path having segment with trailing whitespace',
        action=_validate.paths, args=[['src/utils  ', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_015',
        name='path having whitespace segment',
        action=_validate.paths, args=[['src/ /utils', 'lib']],
        exception=AutopypathError),
    PytestAction('PATHS_016',
        name='Path having segment being only whitespace',
        action=_validate.paths, args=[[Path('src/   /utils'), Path('lib')]],
        exception=AutopypathError),
    PytestAction('PATHS_017',
        name='Path having segment with leading whitespace',
        action=_validate.paths, args=[[Path('src/  utils'), Path('lib')]],
        exception=AutopypathError),
    PytestAction('PATHS_018',
        name='Path having segment with trailing whitespace',
        action=_validate.paths, args=[[Path('src/utils  /more'), Path('lib')]],
        exception=AutopypathError),
    PytestAction('PATHS_019',
        name='Create config with path being empty string',
        action=_validate.paths, args=[['']],
        exception=AutopypathError),
    PytestAction('PATHS_020',
        name='Create config with path being neither str nor Path',
        action=_validate.paths, args=[[123]],
        exception=AutopypathError),
    PytestAction('PATHS_021',
        name='Create config with paths being neither sequence nor None',
        action=_validate.paths, args=[123],
        exception=AutopypathError),
    PytestAction('PATHS_022',
        name='Validate paths attr returns tuple of Path objects',
        action=_validate.paths, args=[['src', 'lib']],
        expected=tuple,
        assertion=Assert.ISINSTANCE),
])
def test_paths(testspec: TestSpec) -> None:
    """Test Config with paths"""
    testspec.run()


@pytest.mark.parametrize('testspec', [
    PytestAction('LOAD_STRATEGY_001',
        name='valid load_strategy',
        action=_validate.load_strategy,
        args=[_LoadStrategy.PREPEND],
        expected=_LoadStrategy.PREPEND),
    PytestAction('LOAD_STRATEGY_002',
        name='load_strategy as string',
        action=_validate.load_strategy,
        args=['prepend'],
        expected=_LoadStrategy.PREPEND),
    PytestAction('LOAD_STRATEGY_003',
        name='invalid load_strategy string',
        action=_validate.load_strategy,
        args=['invalid_strategy'],
        exception=AutopypathError),
    PytestAction('LOAD_STRATEGY_004',
        name='invalid load_strategy type',
        action=_validate.load_strategy,
        args=[123],
        exception=AutopypathError),
    PytestAction('LOAD_STRATEGY_005',
        name='None load_strategy',
        action=_validate.load_strategy,
        args=[None],
        expected=None),
])
def test_load_strategy(testspec: TestSpec) -> None:
    """Test Config with load_strategy"""
    testspec.run()

@pytest.mark.parametrize('testspec', [
    PytestAction('RESOLVE_001',
        name='valid path_resolution_order',
        action=_validate.path_resolution_order,
        args=[['manual', 'pyproject']],
        expected=(_PathResolution.MANUAL, _PathResolution.PYPROJECT)),
    PytestAction('RESOLVE_002',
        name='path_resolution_order as enums',
        action=_validate.path_resolution_order,
        args=[[_PathResolution.PYPROJECT, _PathResolution.MANUAL]],
        expected=(_PathResolution.PYPROJECT, _PathResolution.MANUAL)),
    PytestAction('RESOLVE_003',
        name='invalid path_resolution_order string',
        action=_validate.path_resolution_order,
        args=[['invalid_source']],
        exception=AutopypathError),
    PytestAction('RESOLVE_004',
        name='invalid path_resolution_order type',
        action=_validate.path_resolution_order,
        args=[[123]],
        exception=AutopypathError),
    PytestAction('RESOLVE_005',
        name='empty path_resolution_order',
        action=_validate.path_resolution_order,
        args=[[]],
        expected=None),
    PytestAction('RESOLVE_006',
        name='None path_resolution_order',
        action=_validate.path_resolution_order,
        args=[None],
        expected=None),
    PytestAction('RESOLVE_007',
        name='no path_resolution_order',
        action=_validate.path_resolution_order,
        kwargs={'value': []},
        expected=None),
    PytestAction('RESOLVE_008',
        name='duplicate path_resolution_order entries',
        action=_validate.path_resolution_order,
        args=[['manual', 'pyproject', 'manual']],
        exception=AutopypathError),
    PytestAction('RESOLVE_009',
        name='single path_resolution_order entry',
        action=_validate.path_resolution_order,
        args=[['pyproject']],
        expected=(_PathResolution.PYPROJECT,)),
    PytestAction('RESOLVE_010',
        name='mixed type path_resolution_order entries',
        action=_validate.path_resolution_order,
        args=[['manual', _PathResolution.PYPROJECT]],
        expected=(_PathResolution.MANUAL, _PathResolution.PYPROJECT)),
    PytestAction('RESOLVE_011',
        name='path_resolution_order having leading/trailing whitespace',
        action=_validate.path_resolution_order,
        args=[[' manual ', 'pyproject']],
        exception=AutopypathError),
    PytestAction('RESOLVE_012',
        name='path_resolution_order being neither sequence nor None',
        action=_validate.path_resolution_order,
        args=[123],
        exception=AutopypathError),
    PytestAction('RESOLVE_013',
        name='path_resolution_order containing non-str/non-enum',
        action=_validate.path_resolution_order,
        args=[['manual', 123]],
        exception=AutopypathError),
    PytestAction('RESOLVE_014',
        name="Create config with path_resolution_order as a string",
        action=_validate.path_resolution_order,
        args=['manual'],
        exception=AutopypathError),
])
def test_path_resolution_order(testspec: TestSpec) -> None:
    """Test Config with path_resolution_order"""
    testspec.run()


@pytest.mark.parametrize('testspec', [
    PytestAction('DRY_RUN_001',
        name='dry_run as True',
        action=_validate.dry_run, args=[True],
        expected=True),
    PytestAction('DRY_RUN_002',
        name='dry_run as False',
        action=_validate.dry_run, args=[False],
        expected=False),
    PytestAction('DRY_RUN_003',
        name='dry_run as string',
        action=_validate.dry_run, args=['true'],
        exception=AutopypathError),
    PytestAction('DRY_RUN_004',
        name='dry_run as None',
        action=_validate.dry_run, args=[None],
        exception=AutopypathError),
])
def test_dry_run(testspec: TestSpec) -> None:
    """Test Config with dry_run."""
    testspec.run()

@pytest.mark.parametrize('testspec', [
    PytestAction('STRICT_001',
        name='strict as True',
        action=_validate.strict, args=[True],
        expected=True),
    PytestAction('STRICT_002',
        name='strict as False',
        action=_validate.strict, args=[False],
        expected=False),
    PytestAction('STRICT_003',
        name='strict as string',
        action=_validate.strict, args=['false'],
        exception=AutopypathError),
    PytestAction('STRICT_004',
        name='strict as None',
        action=_validate.strict, args=[None],
        exception=AutopypathError),
])
def test_strict(testspec: TestSpec) -> None:
    """Test Config with strict."""
    testspec.run()

@pytest.mark.parametrize('testspec', [
    PytestAction('LOG_LEVEL_001',
        name='log_level NOTSET',
        action=_validate.log_level, args=[logging.NOTSET],
        expected=logging.NOTSET),
    PytestAction('LOG_LEVEL_002',
        name='log_level DEBUG',
        action=_validate.log_level, args=[logging.DEBUG],
        expected=logging.DEBUG),
    PytestAction('LOG_LEVEL_003',
        name='log_level INFO',
        action=_validate.log_level, args=[logging.INFO],
        expected=logging.INFO),
    PytestAction('LOG_LEVEL_004',
        name='log_level WARNING',
        action=_validate.log_level, args=[logging.WARNING],
        expected=logging.WARNING),
    PytestAction('LOG_LEVEL_005',
        name='log_level ERROR',
        action=_validate.log_level, args=[logging.ERROR],
        expected=logging.ERROR),
    PytestAction('LOG_LEVEL_006',
        name='log_level CRITICAL',
        action=_validate.log_level, args=[logging.CRITICAL],
        expected=logging.CRITICAL),
    PytestAction('LOG_LEVEL_007',
        name='log_level as None (should return NOTSET)',
        action=_validate.log_level, args=[None],
        expected=logging.NOTSET),
    PytestAction('LOG_LEVEL_008',
        name='invalid log_level type (string)',
        action=_validate.log_level, args=['INFO'],
        exception=AutopypathError),
    PytestAction('LOG_LEVEL_009',
        name='invalid log_level type (float)',
        action=_validate.log_level, args=[10.5],
        exception=AutopypathError),
    PytestAction('LOG_LEVEL_010',
        name='invalid log_level value (out of range)',
        action=_validate.log_level, args=[999],
        exception=AutopypathError),
])
def test_log_level(testspec: TestSpec) -> None:
    """Test Config with log_level."""
    testspec.run()

@pytest.mark.parametrize('testspec', [
    PytestAction('PATH_OR_STR_001',
        name='valid string path',
        action=_validate.validate_path_or_str,
        args=['some/path/to/file.txt'],
        expected=Path('some/path/to/file.txt')),
    PytestAction('PATH_OR_STR_002',
        name='valid Path object',
        action=_validate.validate_path_or_str,
        args=[Path('another/path/to/dir')],
        expected=Path('another/path/to/dir')),
    PytestAction('PATH_OR_STR_003',
        name='invalid type (int)',
        action=_validate.validate_path_or_str,
        args=[123],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_004',
        name='path with null byte',
        action=_validate.validate_path_or_str,
        args=['invalid\0path'],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_005',
        name='path that is only forward slashes',
        action=_validate.validate_path_or_str,
        args=['///'],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_006',
        name='path that is only backslashes',
        action=_validate.validate_path_or_str,
        args=['\\\\\\'],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_007',
        name='windows drive letter segment',
        action=_validate.validate_path_or_str,
        args=['C:/folder/file.txt'],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_008',
        name='absolute windows path',
        action=_validate.validate_path_or_str,
        args=['C:\\folder\\subfolder\\file.txt'],
        exception=AutopypathError),
    PytestAction('PATH_OR_STR_009',
        name='absolute posix path',
        action=_validate.validate_path_or_str,
        args=['/folder/subfolder/file.txt'],
        exception=AutopypathError),
])
def test_validate_path_or_str_posix(testspec: TestSpec) -> None:
    """Test _validate_path_or_str function using PytestAction style for posix paths."""
    if os.name != 'posix':
        pytest.skip('Skipping POSIX-specific tests on non-POSIX platform')
    testspec.run()



import ast
from enum import Enum
from unittest.mock import patch

from autopypath._doc_utils import enum_docstrings


@enum_docstrings
class Color(Enum):
    """Enumeration of colors."""

    RED = 1
    """The color red."""

    GREEN = 2
    """The color green."""

    BLUE = 3
    """The color blue."""


def test_enum_docstrings() -> None:
    assert Color.__doc__ == 'Enumeration of colors.'
    assert Color.RED.__doc__ == 'The color red.'
    assert Color.GREEN.__doc__ == 'The color green.'
    assert Color.BLUE.__doc__ == 'The color blue.'


def test_not_an_enum() -> None:
    class NotAnEnum:
        pass

    try:
        enum_docstrings(NotAnEnum)  # type: ignore
    except TypeError as e:
        assert str(e) == 'enum_docstrings can only be applied to Enum subclasses.'
    else:
        raise AssertionError('TypeError was not raised for non-enum class')


def test_enum_docstrings_oserror() -> None:
    class WierdEnum(Enum):
        pass

    with patch('inspect.getsource', side_effect=OSError):
        result = enum_docstrings(WierdEnum)
        assert result is WierdEnum


def test_enum_docstrings_no_classdef() -> None:
    class DummyEnum(Enum):
        X = 1

    # Patch ast.parse to return a module with a non-classdef as the first body element
    fake_mod = ast.Module(body=[ast.Expr(value=ast.Constant(value=42))], type_ignores=[])
    with patch('ast.parse', return_value=fake_mod):
        result = enum_docstrings(DummyEnum)
        assert result is DummyEnum


def test_enum_docstrings_no_docstring() -> None:
    class DocEnum(Enum):
        A = 1
        """docstring for A"""
        B = 2
        """docstring for B"""

    result = enum_docstrings(DocEnum)
    assert result is DocEnum, f'Expected DocEnum, got {result!r}'
    assert DocEnum.A.__doc__ == 'docstring for A', f'Expected docstring for A, got {DocEnum.A.__doc__!r}'
    assert DocEnum.B.__doc__ == 'docstring for B', f'Expected docstring for B, got {DocEnum.B.__doc__!r}'


def test_enum_docstrings_unassigned_docstring() -> None:
    class UnassignedEnum(Enum):
        """An enum with unassigned docstrings"""

        X = 1
        """Unassigned docstring for X"""

        Y = 2
        """Unassigned docstring for Y"""

    unassigned = 'An enum with unassigned docstrings'
    result = UnassignedEnum
    assert result is UnassignedEnum
    assert UnassignedEnum.X.__doc__ == unassigned, f'Expected {unassigned!r}, got {UnassignedEnum.X.__doc__!r}'
    assert UnassignedEnum.Y.__doc__ == unassigned, f'Expected {unassigned!r}, got {UnassignedEnum.Y.__doc__!r}'


def test_enum_docstrings_mixed() -> None:
    @enum_docstrings
    class MixedEnum(Enum):
        """An enum with mixed docstrings"""

        A = 1
        """Docstring for A"""

        B = 2
        # No docstring for B

        C = 3
        """Docstring for C"""

    unassigned = 'An enum with mixed docstrings'

    result = MixedEnum
    assert result is MixedEnum
    assert MixedEnum.A.__doc__ == 'Docstring for A', f'Expected docstring for A, got {MixedEnum.A.__doc__!r}'
    assert MixedEnum.B.__doc__ == unassigned, f'Expected {unassigned!r}, got {MixedEnum.B.__doc__!r}'
    assert MixedEnum.C.__doc__ == 'Docstring for C', f'Expected docstring for C, got {MixedEnum.C.__doc__!r}'


class DummyEnum(Enum):
    A = 1


def test_enum_docstrings_invalid_ast() -> None:
    """Test enum_docstrings when AST does not contain a ClassDef."""
    # Patch inspect.getsource to return a function definition instead of a class
    fake_source = 'def not_a_class():\n    pass\n'
    with patch('inspect.getsource', return_value=fake_source):
        result = enum_docstrings(DummyEnum)
        assert result is DummyEnum  # Should return the enum unchanged



"""Config from a toml file for autopypath.


It handles loading and parsing configuration options from a specified toml file,
such as `pyproject.toml` or `autopypath.toml`. The configuration options include
repository markers, additional paths to include in PYTHONPATH, load strategy,
and path resolution order.

This module defines the :class:`_TomlConfig` class, which extends the
base :class:`_Config` class to provide toml-specific configuration loading.


The toml configuration is expected to be in a specific section of the toml file,
for example, `[tool.autopypath]` in `pyproject.toml`. The class reads this section
and extracts the relevant configuration options, validating their types and values.

The formats for the configuration options in the toml file are as follows:

.. code-block:: toml
    [tool.autopypath]
    repo_markers = { '.git' = 'dir', 'setup.py' = 'file' }
    paths = ['src', 'tests']
    load_strategy = 'prepend'
    path_resolution_order = ['manual', 'pyproject']

"""

import datetime
from pathlib import Path
from types import MappingProxyType
from typing import Any, Union

import tomli

from ... import _validate
from ..._exceptions import AutopypathError
from ..._load_strategy import _LoadStrategy
from ..._log import _log
from ..._marker_type import _MarkerType
from ..._path_resolution import _PathResolution
from ..._types import _NoPath
from ._config import _Config

__all__ = ['_TomlConfig']


_TOML_TYPES: dict[type, str] = {
    dict: 'table',
    list: 'array',
    str: 'string',
    int: 'int',
    float: 'float',
    bool: 'boolean',
    datetime.date: 'date',
    datetime.datetime: 'datetime',
    datetime.time: 'time',
}
"""Mapping of Python types to TOML types for error messages."""


class _TomlConfig(_Config):
    """Configuration for autopypath using toml files."""

    __slots__ = ('_repo_root_path', '_toml_filepath', '_toml_section', '_no_file_found')

    def __init__(self, *, repo_root_path: Union[Path, None], toml_filename: str, toml_section: str) -> None:
        """Configuration for autopypath using toml files.

        :param Path | None repo_root_path: The root path of the repository containing the toml file.
            If ``None``, a special empty configuration is created with the :attr:`toml_filepath` property
            set to a :class:`NoPath` instance (a custom Path subclass representing the absence of a path).
            The configuration attributes (:attr:`repo_markers`, :attr:`paths`, :attr:`load_strategy`, and
            :attr:`path_resolution_order`) will all be set to ``None`` in this case.
        :param str toml_filename: The name of the toml file to load (e.g., 'pyproject.toml', 'autopypath.toml').
        :param str toml_section: The section in the toml file to load (e.g., 'tool.autopypath').
        :raises AutopypathError: If any configuration value is of an invalid type (ex. string instead of table).
        :raises AutopypathError: If any configuration value is invalid.
        :raises FileNotFoundError: If the toml file does not exist.
        """
        self._repo_root_path: Path = _NoPath()
        self._toml_filepath: Path = _NoPath()
        self._no_file_found: bool = False  # Indicates if the toml file was not found
        self._toml_section: str = _validate.toml_section(toml_section)

        if repo_root_path is None:
            super().__init__(repo_markers=None, paths=None, load_strategy=None, path_resolution_order=None)
            return
        self._repo_root_path = _validate.root_repo_path(repo_root_path)
        self._toml_filepath = _validate.toml_filename(toml_filename)

        # toml data
        toml_data = self._toml_data()
        if not toml_data:
            super().__init__(repo_markers=None, paths=None, load_strategy=None, path_resolution_order=None)
            return

        # [tool.autopypath]
        autopypath_config = self._toml_autopypath(toml_data)
        if not autopypath_config:
            super().__init__(repo_markers=None, paths=None, load_strategy=None, path_resolution_order=None)
            return

        # example: repo_markers = { '.git' = 'dir', 'setup.py' = 'file' }
        repo_markers = self._toml_repo_markers(autopypath_config)

        # example: paths = ['src', 'tests']
        paths = self._toml_paths(autopypath_config)
        # example: load_strategy = 'prepend'
        load_strategy = self._toml_load_strategy(autopypath_config)

        # example: path_resolution_order = ['manual', 'pyproject']
        path_resolution_order = self._toml_path_resolution_order(autopypath_config)

        super().__init__(
            repo_markers=repo_markers,
            paths=paths,
            load_strategy=load_strategy,
            path_resolution_order=path_resolution_order,
        )

    @property
    def toml_filepath(self) -> Path:
        """Returns the toml file path used for this configuration.

        :return Path: The toml file path.
        """
        return self._toml_filepath

    @property
    def toml_section(self) -> str:
        """Returns the toml section used for this configuration.

        :return str: The toml section.
        """
        return self._toml_section

    @property
    def no_file_found(self) -> bool:
        """Indicates whether the toml file was not found.

        :return bool: ``True`` if the toml file does not exist, ``False`` otherwise.
        """
        toml_path = self._repo_root_path / self._toml_filepath
        return not toml_path.exists() or not toml_path.is_file()

    def _toml_data(self) -> dict[str, Any]:
        """Loads and returns the toml data as a dictionary.

        :return dict[str, Any]: The parsed toml data or an empty dictionary if the file does not exist.
        """
        toml_path = self._repo_root_path / self._toml_filepath
        if not toml_path.exists() or not toml_path.is_file():
            self._no_file_found = True
            _log.debug(f'No {self._toml_filepath} file found at {toml_path}.')
            return {}

        try:
            with toml_path.open('rb') as f:
                toml_data = tomli.load(f)
        except tomli.TOMLDecodeError as e:
            raise AutopypathError(f'Error parsing {toml_path}: {e}') from e
        return toml_data

    def _toml_autopypath(self, toml_data: dict[str, Any]) -> dict[str, Any]:
        """Returns the [tool.autopypath] section from pyproject.toml
        if it exists, otherwise returns an empty dictionary.


        Example
        -------
        .. code-block:: toml
            [tool.autopypath]
            repo_markers = { '.git' = 'dir', 'setup.py' = 'file' }
            paths = ['src', 'tests']
            path_resolution_order = ['manual', 'pyproject']

        :param dict[str, Any] pyproject_data: The parsed pyproject.toml data.
        :return dict[str, Any]: The autopypath configuration section.
        :raises AutopypathError: If the configuration for [tool.autopypath] is invalid.
        """
        autopypath_config = toml_data.get('tool', {}).get('autopypath', {})
        if not isinstance(autopypath_config, dict):
            toml_type = _TOML_TYPES.get(type(autopypath_config), 'unknown')
            raise AutopypathError(
                'Invalid [tool.autopypath] section in pyproject.toml: expected table, '
                f'got {toml_type}: {autopypath_config}'
            )
        if not autopypath_config:
            _log.debug('No [tool.autopypath] configuration section found in pyproject.toml.')
            return {}

        return autopypath_config

    def _toml_repo_markers(self, autopypath_config: dict) -> Union[MappingProxyType[str, _MarkerType], None]:
        """Collects repository markers from the [tool.autopypath] section of the toml data.

        Example
        -------
        .. code-block:: toml
            [tool.autopypath]
            repo_markers = { 'custom_marker.txt' = 'file', '.git' = 'dir' }

        :param autopypath_config: The autopypath configuration section from the toml data.
        :return MappingProxyType[str, MarkerType] | None: A mapping of repository markers or ``None``.
        :raises AutopypathError: If the configuration is invalid.
        :raises AutopypathError: If any marker type is invalid (not 'file' or 'dir').
        """
        raw_repo_markers = autopypath_config.get('repo_markers', None)
        if not isinstance(raw_repo_markers, (type(None), dict)):
            toml_type = _TOML_TYPES.get(type(raw_repo_markers), 'unknown')
            raise AutopypathError(
                f'Invalid repo_markers in pyproject.toml: expected table, got {toml_type}: {raw_repo_markers}'
            )
        collected_repo_markers: dict[str, _MarkerType] = {}
        if raw_repo_markers is not None:
            for key, value in raw_repo_markers.items():
                if not isinstance(value, str):
                    toml_type = _TOML_TYPES.get(type(raw_repo_markers), 'unknown')
                    raise AutopypathError(
                        f'Invalid repo_marker value for {key}: expected string, got {toml_type}: {value}')
                try:
                    collected_repo_markers[key] = _MarkerType(value)
                except ValueError as e:
                    raise AutopypathError(
                        f'Invalid repo_marker value for {key}: expected "file" or "dir", got {value}'
                    ) from e

        repo_markers = _validate.repo_markers(collected_repo_markers)
        return repo_markers

    def _toml_paths(self, autopypath_config: dict[str, Any]) -> Union[tuple[Path, ...], None]:
        """Collects paths from toml configuration.

        Example
        -------
        .. code-block:: toml
            [tool.autopypath]
            paths = ['src', 'tests']

        :param autopypath_config: The autopypath configuration section from toml.
        :return tuple[Path, ...] | None: A tuple of paths or ``None``.
        :raises AutopypathError: If the configuration is invalid.
        """
        # example: paths = ['src', 'tests']
        raw_paths = autopypath_config.get('paths', None)
        if not isinstance(raw_paths, (type(None), list)):
            toml_type = _TOML_TYPES.get(type(raw_paths), 'unknown')
            raise AutopypathError(f'Invalid paths in {self._toml_filepath}: '
                                  f'expected array, got {toml_type}: {raw_paths}')
        requested_paths = _validate.paths(raw_paths)
        filtered_paths: list[Path] = []
        if requested_paths is not None:
            for p in requested_paths:
                target_path = self._repo_root_path / p
                filtered_paths.append(target_path)
        paths = tuple(filtered_paths) if filtered_paths else None
        return paths

    def _toml_load_strategy(self, autopypath_config: dict[str, Any]) -> Union[_LoadStrategy, None]:
        """Collects load strategy from toml configuration.

        Example
        -------
        .. code-block:: toml
            [tool.autopypath]
            load_strategy = 'append'

        :param autopypath_config: The autopypath configuration section from toml.
        :return LoadStrategy | None: The load strategy or ``None``.
        :raises AutopypathError: If the configuration is invalid.
        """
        raw_load_strategy = autopypath_config.get('load_strategy', None)
        if not isinstance(raw_load_strategy, (type(None), str)):
            toml_type = _TOML_TYPES.get(type(raw_load_strategy), 'unknown')
            raise AutopypathError(
                f'Invalid load_strategy in {self._toml_filepath}: expected string, got {toml_type}: {raw_load_strategy}'
            )
        if raw_load_strategy is None:
            return None
        try:
            load_strategy = _LoadStrategy(raw_load_strategy)
        except ValueError as e:
            raise AutopypathError(f'Invalid load_strategy: {raw_load_strategy}') from e

        return load_strategy

    def _toml_path_resolution_order(
            self, autopypath_config: dict[str, Any]) -> Union[tuple[_PathResolution, ...], None]:
        """Collects path resolution order from toml configuration.

        Example
        -------
        .. code-block:: toml
            [tool.autopypath]
            path_resolution_order = ['manual', 'pyproject']

        :param autopypath_config: The autopypath configuration section from toml.
        :return tuple[str, ...] | None: The path resolution order or ``None``.
        :raises AutopypathError: If the configuration is invalid.
        """
        raw_order = autopypath_config.get('path_resolution_order', None)
        if not isinstance(raw_order, (type(None), list)):
            toml_type = _TOML_TYPES.get(type(raw_order), 'unknown')
            raise AutopypathError(
                f'Invalid path_resolution_order in {self._toml_filepath}: expected array, got {toml_type}: {raw_order}'
            )
        path_resolution_order = _validate.path_resolution_order(raw_order)
        return path_resolution_order

    def __repr__(self) -> str:
        """String representation of the TomlConfig object.

        :return str: A string representation of the TomlConfig instance.
        """
        repo_root_str = 'None' if isinstance(self._repo_root_path, _NoPath) else f'{self._repo_root_path!r}'
        return (
            f'{self.__class__.__name__}(repo_root_path={repo_root_str}, '
            f'toml_filename={str(self._toml_filepath)!r}, '
            f'toml_section={self._toml_section!r})\n'
            f'#  repo_markers={self.repo_markers!r}\n'
            f'#  paths={self.paths!r}\n'
            f'#  load_strategy={self.load_strategy!r}\n'
            f'#  path_resolution_order={self.path_resolution_order!r}'
        )

    def __str__(self) -> str:
        """String representation of the TomlConfig instance."""
        return (
            f'{self.__class__.__name__}:\n'
            f'  repo_markers={self.repo_markers!r}\n'
            f'  paths={self.paths!r}\n'
            f'  load_strategy={self.load_strategy!r}\n'
            f'  path_resolution_order={self.path_resolution_order!r}'
        )



"""Validators for configuration parameters.

Use by importing the module and calling the desired validation function.

The validation functions raise `AutopypathError` if the input
is invalid.

They return the validated and possibly transformed value if valid."""
import re
from collections.abc import Mapping, Sequence
from ntpath import pathsep as nt_pathsep
from os import sep as path_sep
from pathlib import Path, PurePosixPath
from posixpath import pathsep as posix_pathsep
from types import MappingProxyType
from typing import Any, Union

from ._exceptions import AutopypathError
from ._load_strategy import _LoadStrategy, resolve_load_strategy_literal
from ._log import _log
from ._marker_type import _MarkerType, resolve_marker_type_literal
from ._path_resolution import _PathResolution, resolve_path_resolution_literal

__all__: Sequence[str] = []  # No exports; functions are used internally.

_MAX_FILE_DIR_NAME_LENGTH: int = 64
"""Maximum length for file or directory names.

Deliberately set to 64 to allow for future flexibility and
avoid issues with filesystems that may have lower limits.
"""


def log_level(value: Any) -> int:
    """Validates the log level.

    If the input is ``None``, returns ``logging.NOTSET``.

    It verifies that the input is an integer and one of the valid logging levels
    defined in the `logging` module.

    :param Any value: The log level to validate.
    :return int: A validated log level.
    :raises AutopypathError: If the input is not an integer or None.
    :raises AutopypathError: If the log level is not a valid logging level.
    """
    import logging

    if value is None:
        return logging.NOTSET

    if not isinstance(value, int):
        raise AutopypathError(f'Invalid log_level: expected int, got {type(value)}')

    valid_levels = {
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    }

    if value not in valid_levels:
        raise AutopypathError(f'Invalid log_level: {value} is not a valid logging level')

    return value


def strict(value: Any) -> bool:
    """Validates the strict parameter.

    :param Any value: The strict value to validate.
    :return bool: A validated boolean value for strict.
    :raises AutopypathError: If the input is not a boolean.
    """
    if not isinstance(value, bool):
        raise AutopypathError(f'Invalid strict: expected bool, got {type(value)}')
    return value


def toml_filename(value: Any) -> Path:
    """Validates the TOML filename.

    :param Any value: The TOML filename to validate.
    :return Path: A validated TOML Path object.
    :raises AutopypathError: If the input is not a string.
    :raises AutopypathError: If the filename is invalid.
    :raises AutopypathError: If the filename does not end with .toml
    """
    if not isinstance(value, str):
        raise AutopypathError(f'Invalid toml_filename: expected str, got {type(value)}')
    validate_file_or_dir_name(value)
    if not value.lower().endswith('.toml'):
        raise AutopypathError(f'Invalid toml_filename: {value!r} does not end with .toml')
    return Path(value)


_TOML_SECTION_RE: re.Pattern[str] = re.compile(r'^[A-Za-z0-9](?:[A-Za-z0-9_.-]*[A-Za-z0-9])?$')
"""Regular expression for validating TOML section names.
- Must start and end with an alphanumeric character.
- Can contain alphanumeric characters, underscores, dashes,  and dots in between.
- Cannot be empty.

Check for adjecent dots, dashes, or underscores is enforced in a second regex.
"""

_ADJACENT_INVALID_SEQUENCES_RE: re.Pattern[str] = re.compile(r'[._-]{2,}')
"""Regular expression for detecting characters that cannot be adjacent in TOML section names.

Does not allow: '.', '-', or '_' to be adjacent to each other or themselves.
"""


def toml_section(value: Any) -> str:
    """Validates the TOML section name.

    :param Any value: The TOML section name to validate.
    :return str: A validated TOML section name.
    :raises AutopypathError: If the input is not a string.
    :raises AutopypathError: If the section name is invalid.
    """
    if not isinstance(value, str):
        raise AutopypathError(f'Invalid toml_section: expected str, got type {type(value)}')
    if value.strip() == '':
        raise AutopypathError('Invalid toml_section: section name cannot be empty')
    if not _TOML_SECTION_RE.match(value):
        raise AutopypathError(f'Invalid toml_section name: {value!r} does not '
                              'match required pattern for toml section names')
    if _ADJACENT_INVALID_SEQUENCES_RE.search(value):
        raise AutopypathError(f'Invalid toml_section name: {value!r} '
                              'cannot have adjacent dots, dashes, or underscores')
    return value


def root_repo_path(value: Any) -> Path:
    """Validates the repository root path.

    :param Any value: The repository root path to validate.
    :return Path: A validated Path object representing the repository root.
    :raises AutopypathError: If the input is not a Path or string.
    """
    repo_path = validate_path_or_str(value)
    if not repo_path.exists() or not repo_path.is_dir():
        raise AutopypathError(f'Repository root path does not exist or is not a directory: {repo_path}')
    return repo_path


def context_file(value: Any) -> Path:
    """Validates the context file path.

    :param Any value: The context file path to validate.
    :return Path: A validated Path object representing the context file.
    :raises AutopypathError: If the input is not a Path or string.
    :raises AutopypathError: If the file does not exist or is not a file.
    :raises AutopypathError: If the path is invalid.
    """
    try:
        if isinstance(value, Path):
            _log.debug('Validating context file path: %s', value)
        elif isinstance(value, str):
            value = validate_path_or_str(value)
            value = Path(value)
        _log.debug('Validating context file path: %s', value)
        if not value.exists() or not value.is_file():
            raise AutopypathError(f'Context file does not exist or is not a file: {value}')
    except AutopypathError:
        raise
    except BaseException as e:
        raise AutopypathError(f'Invalid context file path: {e}') from e

    return value


def repo_markers(value: Any) -> Union[MappingProxyType[str, _MarkerType], None]:
    """Validates a mapping of repository markers.

    :param Any value: A mapping where keys are filenames or directory names
        that indicate the repository root, and values are of type `MarkerType` or strings.
    :return MappingProxyType[str, MarkerType] | None: A validated immutable mapping of repository markers.
        or None if the input is None.
    :raises AutopypathError: If the input is not a mapping, keys are not strings,
        or values are not strings.
    :raises AutopypathError: If any value is not a valid `MarkerType`.
    """
    if value is None:
        return None

    if not isinstance(value, Mapping):
        raise AutopypathError(f'Invalid repo_markers: expected a mapping, got {type(value)}')
    validated_markers: dict[str, _MarkerType] = {}
    for key, val in value.items():
        if isinstance(val, str):
            resolved_val = resolve_marker_type_literal(val)
            if resolved_val is None:
                raise AutopypathError(f'Invalid MarkerType: {val} is not a valid MarkerType')
            val = resolved_val
        if not isinstance(val, _MarkerType):
            raise AutopypathError(f'Invalid repo_markers value: expected MarkerType or string, got {type(val)}')
        if not isinstance(key, str):
            raise AutopypathError(f'Invalid repo_markers key: expected str, got {type(key)}')
        validate_file_or_dir_name(key)
        validated_markers[key] = val

    if len(validated_markers) == 0:
        return None

    return MappingProxyType(validated_markers)


def paths(value: Any) -> Union[tuple[Path, ...], None]:
    """Validates a sequence of Path objects or strings.

    Strings are ALWAYS treated as POSIX paths for consistency, regardless of the operating system.

    If the input is None, returns None. If the input is a sequence, each item is validated
    to be either a Path or a string (which is converted to a Path).

    If the sequence is empty, returns ``None``

    :param Any value: A sequence of Path objects or strings.
    :return tuple[Path, ...] | None: A validated tuple of :class:`Path` objects or None if the input is None.
    :raises AutopypathError: If the input is not a sequence or contains non-Path or non-str items
    """
    if value is None:
        return None

    if not isinstance(value, Sequence):
        raise AutopypathError(f'Invalid paths: expected a sequence, got {type(value)}')

    if len(value) == 0:
        return None

    validated_paths: list[Path] = []
    for item in value:
        validated_paths.append(validate_path_or_str(item))
    return tuple(validated_paths)


def load_strategy(value: Any) -> Union[_LoadStrategy, None]:
    """Validates a LoadStrategy value.

    :param Any value: A LoadStrategy value or string matching a LoadStrategy value.
    :return LoadStrategy | None: A validated LoadStrategy value or None if the input is None.
    :raises AutopypathError: If the input string does not match any LoadStrategy value
    :raises AutopypathError: If the input is not a LoadStrategy or a string.
    """
    if value is None:
        return None

    if isinstance(value, str):
        resolved_value = resolve_load_strategy_literal(value)
        if resolved_value is None:
            raise AutopypathError(f'Invalid LoadStrategy: {value} is not a valid LoadStrategy: {list(_LoadStrategy)}')
        value = resolved_value

    if not isinstance(value, _LoadStrategy):
        raise AutopypathError(f'Invalid load_strategy: expected LoadStrategy, got {type(value)}')
    return value


def path_resolution_order(value: Any) -> Union[tuple[_PathResolution, ...], None]:
    """Validates a sequence of PathResolution values.

    :param Any value: A sequence of PathResolution values or strings matching PathResolution values.
    :return tuple[PathResolution, ...] | None: A validated sequence of PathResolution values or None
        if the input is None or empty.
    :raises AutopypathError: If the input is not a sequence or contains non-PathResolution items
    :raises AutopypathError: If there are duplicate PathResolution values
    """
    if value is None:
        return None

    if isinstance(value, (str, bytes)):
        raise AutopypathError(f'Invalid path_resolution_order: expected a sequence, got {type(value)}')

    if not isinstance(value, Sequence):
        raise AutopypathError(f'Invalid path_resolution_order: expected a sequence, got {type(value)}')

    if len(value) == 0:
        return None

    validated_orders: list[_PathResolution] = []
    seen_orders: dict[_PathResolution, int] = {}
    for item in value:
        if isinstance(item, str):
            resolved_item = resolve_path_resolution_literal(item)
            if resolved_item is None:
                raise AutopypathError(
                    f'Invalid PathResolution: {item} is not a valid PathResolution: {list(_PathResolution)}'
                )
            item = resolved_item
        if not isinstance(item, _PathResolution):
            raise AutopypathError(f'Invalid path_resolution_order item: expected PathResolution, got {item}')
        validated_orders.append(item)
        seen_orders[item] = seen_orders.get(item, 0) + 1
    if len(seen_orders) != len(validated_orders):
        duplicates = set(order for order in validated_orders if seen_orders[order] > 1)
        raise AutopypathError(f'Duplicate PathResolution values are not allowed: Duplicated {duplicates}')

    return tuple(validated_orders)

def _normalize_path_string_to_platform(raw_path_string: str) -> Path:
    """
    Normalize a POSIX style path string to the current platform's Path object.

    1. Syntactic Gatekeeping: No backslashes allowed in the config string.
    2. POSIX Validation: Standard check for leading '/'
    3. Windows Leakage Validation (The "C:" Check):
       We must explicitly check for colons in the first part of the path.
       In POSIX, "C:/" is relative; on Windows, it is absolute. We reject it.
    4. Host Translation: Direct wrapper to the platform-specific Path.

    :raise AutopypathError: If the path string contains backslashes, is absolute in POSIX,
        or uses Windows drive/volume syntax.
    :param str raw_path_string: The raw POSIX style path string to normalize.
    :return Path: A Path object representing the normalized path on the current platform.
    """
    # 1. Syntactic Gatekeeping: No backslashes allowed in the config string.
    if "\\" in raw_path_string:
        raise AutopypathError("backslashes forbidden in path string.")

    # 2. POSIX Validation: Standard check for leading '/'
    posix_path = PurePosixPath(raw_path_string)
    if posix_path.is_absolute():
        raise AutopypathError("POSIX absolute paths forbidden.")

    # 3. Windows Leakage Validation (The "C:" Check):
    # We must explicitly check for colons in the first part of the path.
    # In POSIX, "C:/" is relative; on Windows, it is absolute. We reject it.
    first_part = posix_path.parts[0] if posix_path.parts else ""
    if ":" in first_part:
        raise AutopypathError("Windows drive/volume syntax forbidden.")

    # 4. Host Translation: Direct wrapper to the platform-specific Path.
    return Path(posix_path)


def validate_path_or_str(path: Union[Path, str]) -> Path:
    """Validate a Path object or a string path.

    It does not check for existence, only validity.

    The returned Path object is the same type as the input if the input is a Path.
    If the input is a string, it is parsed as a POSIX path and returned
    as the current platform's :class:`Path` version.

    - Cannot contain null bytes.
    - Cannot be empty or whitespace only.
    - Cannot have leading or trailing whitespace.
    - Cannot be only backslashes or only forward slashes.
    - Each segment cannot be an invalid file or directory name.

    :param Path | str path: The Path object or string path to validate.
    :raises AutopypathError: If the path is invalid.
    :raises AutopypathError: If the input is not a Path or string.
    :return Path: The validated Path object.
    """
    if not isinstance(path, (Path, str)):
        raise AutopypathError(
            f'Invalid path: expected Path or str, got {type(path)}')
    item_str: str = str(path) if isinstance(path, Path) else path
    _log.debug('Validating path: %s', item_str)
    if '\000' in item_str:
        raise AutopypathError(
            'Invalid path item: path cannot contain null byte')
    if item_str.strip() == '':
        raise AutopypathError(
            'Invalid path item: path cannot be empty or only whitespace')
    if item_str.lstrip() != item_str:
        raise AutopypathError(
            'Invalid path item: path cannot have leading whitespace ')
    if item_str.rstrip() != item_str:
        raise AutopypathError(
            'Invalid path item: path cannot have trailing whitespace')
    if item_str.replace('\\', '') == '':
        raise AutopypathError(
            'Invalid path item: path cannot be only backslashes')
    if item_str.replace('/', '') == '':
        raise AutopypathError(
            'Invalid path item: path cannot be only forward slashes')
    validated_path = _normalize_path_string_to_platform(
        item_str) if isinstance(path, str) else path

    # Code here will work on any OS, but won't be fully tested on a single OS.
    # On POSIX platforms the root '/' is skipped. On Windows the drive letter root is skipped.
    # There is a need to write a monkeypatched test to fully cover this from a
    # single OS.
    for offset, segment in enumerate(validated_path.parts):
        if offset == 0 and segment == '/':
            # Skip root '/' for POSIX even on non-POSIX platforms
            continue
        if offset == 0 and (re.match(r'^[A-Za-z]:$', segment)
                            or re.match(r'^[A-Za-z]:\\$', segment)):
            # Skip Windows drive letter root even on non-Windows platforms
            continue
        validate_file_or_dir_name(segment)

    return validated_path


def validate_file_or_dir_name(name: str) -> None:
    """Check if a given name is forbidden as a file or directory name.

    - Cannot be empty or whitespace only.
    - Cannot have leading or trailing whitespace.
    - Cannot contain path separators.
    - Cannot contain forbidden characters for file/directory names
      (e.g., `<`, `>`, `:`, `"`, `/`, `\\`, `|`, `?`, `*` on Windows
      and `/` on POSIX systems). Characters forbidden on either system are
      always forbidden.
    - Cannot be a reserved name for Windows.
    - Cannot exceed 64 characters in length.

    :param str name: The file or directory name to validate.
    :raises AutopypathError: If the name is invalid.

    """
    _log.debug('Validating file or directory name: %s', name)
    if name.strip() == '':
        raise AutopypathError(f'Invalid file/dir name: cannot be empty or whitespace: {name!r}')
    if name.lstrip() != name:
        raise AutopypathError(f'Invalid file/dir name: cannot have leading whitespace: {name!r}')
    if name.rstrip() != name:
        raise AutopypathError(f'Invalid file/dir name: cannot have trailing whitespace: {name!r}')
    if posix_pathsep in name or nt_pathsep in name or path_sep in name:
        raise AutopypathError(f'Invalid file/dir name: cannot contain path separators: {name!r}')
    if has_forbidden_chars(name) or is_windows_reserved(name):
        raise AutopypathError(f'Invalid file/dir name: {name!r} is not allowed')
    if len(name) > _MAX_FILE_DIR_NAME_LENGTH:
        message = f'Invalid file/dir name: {name!r} exceeds maximum length of {_MAX_FILE_DIR_NAME_LENGTH} characters'
        raise AutopypathError(message)


def has_forbidden_chars(name: str) -> bool:
    """Check if a given name contains forbidden characters for file or directory names.

    :param str name: The file or directory name to check.
    :return bool: True if the name contains forbidden characters, False otherwise.
    """
    forbidden = set('<>:"/\\|?*\0')  # Common forbidden characters on Windows and POSIX
    return any(c in name for c in forbidden)


def is_windows_reserved(name: str) -> bool:
    """Check if a given name is a reserved name on Windows.

    :param str name: The file or directory name to check.
    :return bool: True if the name is a reserved name on Windows, False otherwise.
    """
    reserved = {
        'CON',
        'PRN',
        'AUX',
        'NUL',
        *(f'COM{i}' for i in range(1, 10)),
        *(f'LPT{i}' for i in range(1, 10)),
    }
    # Windows ignores case and extension for reserved names
    base = name.split('.')[0].upper()
    return base in reserved


def dry_run(value: Any) -> bool:
    """Validates the dry_run parameter.

    :param Any value: The dry_run value to validate.
    :return bool: A validated boolean value for dry_run.
    :raises AutopypathError: If the input is not a boolean.
    """
    if not isinstance(value, bool):
        raise AutopypathError(f'Invalid dry_run: expected bool, got {type(value)}')
    return value



"""TestSpec testing framework - context management."""

from typing import Any


class Context(dict[str, Any]):
    """A simple context dictionary for holding test parameters and state.

    This class extends the built-in dict to provide a convenient way to
    store and access context-specific data during test execution.

    It can be used to share data between different parts of a test case,
    such as setup, execution, and validation phases.
    """

    def __init__(self, /, **kwargs: Any) -> None:
        """Initialize the Context with optional keyword arguments.

        Keys must be valid python identifiers (str) and cannot shadow existing attributes or methods.

        Example:

            context = Context(param1=42, param2="value")

        :param kwargs: Key-value pairs to initialize the context dictionary.
            They must be valid python identifiers as keys (str) and can have any type as values.
        :type kwargs: Any
        :raises TypeError:
            - If any key in kwargs is not of type str.
            - If any key in kwargs is not a valid python identifier (str).
            - If any key in kwargs shadows existing attributes or methods of the Context class.
            - If any key in kwargs starts or ends with double underscores '__'.
        """
        # pylint: disable=consider-iterating-dictionary
        if not all(isinstance(key, str) for key in kwargs.keys()):
            raise TypeError('All keys in Context must be of type str')
        if not all(key.isidentifier() for key in kwargs.keys()):
            raise TypeError('All keys in Context must be valid python identifiers')
        if not all(not key.startswith('__') and not key.endswith('__') for key in kwargs.keys()):
            raise TypeError("Context keys cannot start or end with double underscores '__'")
        if not all(not hasattr(self, key) for key in kwargs.keys()):
            raise TypeError('Context keys cannot shadow existing attributes or methods')
        super().__init__(**kwargs)

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the context dictionary.

        :param name: The name of the attribute to get.
        :type name: str
        :return: The value of the attribute.
        :rtype: Any
        :raises AttributeError: If the attribute is not found.
        """
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(f"'Context' object has no attribute '{name}'") from e  # pylint: disable=raise-missing-from



"""Resolution order definitions for autopypath."""

from enum import Enum
from types import MappingProxyType
from typing import Union

from ._typing import Final, Literal, TypeAlias, TypeGuard

__all__ = ['_PathResolution']


class _PathResolution(str, Enum):
    """Defines the order in which :data:`sys.path` sources are resolved.

    - MANUAL: Paths provided directly to the configuration function.
    - AUTOPYPATH: Paths specified in a `autopypath.toml` file.
    - PYPROJECT: Paths specified in `pyproject.toml` in the repository root.

    Example
    -------
    .. code-block:: python
        from autopypath.path_resolution import PathResolution

        order = PathResolution.ENV
    """

    MANUAL = 'manual'
    """Paths provided directly via the `paths` parameter to `configure_pypath()`."""
    AUTOPYPATH = 'autopypath'
    """Paths specified in a `autopypath.toml` file."""
    PYPROJECT = 'pyproject'
    """Paths specified in the [tool.autopypath] section of the `pyproject.toml` file in the repository root."""



PathResolutionLiteral: TypeAlias = Literal['manual', 'autopypath', 'pyproject']
"""Literal type for PathResolution values."""


PATH_RESOLUTION_MAP: Final[MappingProxyType[PathResolutionLiteral, _PathResolution]] = MappingProxyType(
    {order.value: order for order in _PathResolution}
)
"""Mapping from literal strings to PathResolution enum members.

Example
-------

.. code-block:: python

    from autopypath.path_resolution import RESOLUTION_ORDER_MAP, PathResolution

    order = RESOLUTION_ORDER_MAP['env']
    assert order == PathResolution.ENV
"""


def is_path_resolution_literal(value: str) -> TypeGuard[PathResolutionLiteral]:
    """Checks if the given string is a valid PathResolution literal.

    Example
    -------

    .. code-block:: python
        from autopypath.path_resolution import is_path_resolution_literal

        assert is_path_resolution_literal('manual') is True
        assert is_path_resolution_literal('autopypath') is True
        assert is_path_resolution_literal('pyproject') is True
        assert is_path_resolution_literal('invalid') is False

    :param str value: The string to check.
    :return bool: ``True`` if the string is a valid ``PathResolution`` literal, ``False`` otherwise.
    """
    return value in PATH_RESOLUTION_MAP


def resolve_path_resolution_literal(value: str) -> Union[_PathResolution, None]:
    """Resolves a string literal to its corresponding PathResolution enum member
    or returns ``None`` if the literal is invalid.

    Example
    -------

    .. code-block:: python
        from autopypath.path_resolution import resolve_literal, PathResolution

        assert resolve_path_resolution_literal('manual') == PathResolution.MANUAL
        assert resolve_path_resolution_literal('autopypath') == PathResolution.AUTOPYPATH
        assert resolve_path_resolution_literal('pyproject') == PathResolution.PYPROJECT
        assert resolve_path_resolution_literal('invalid') is None

    :param str value: The string literal to resolve.
    :return PathResolution | None: The corresponding PathResolution enum member,
                                    or ``None`` if the literal is invalid.
    """
    if is_path_resolution_literal(value):
        return PATH_RESOLUTION_MAP.get(value)
    return None



"""TestSpec testing framework - tagged exceptions."""
from enum import Enum
from typing import Any, Generic, TypeVar

E = TypeVar('E', bound=Exception)


class TaggedException(Exception, Generic[E]):
    """
    A generic exception that can be specialized with a base exception type
    and requires a tag during instantiation.

    This helper class extends the built-in Exception class and adds a mandatory tag
    attribute usable by the testspec framework. The testspec framework uses this class
    to create exceptions with specific tags for error handling and identification.

    Tagged exceptions are intended to provide additional context or categorization
    for the exception when testing but are not mandatory when testing exceptions.

    To create a tagged exception, subclass this class with the desired base exception type.
    The base exception type must be a subclass of Exception.
    The subclassed tagged exception can then be raised like a normal exception,

    The tag must be an instance of Enum to ensure a controlled set of possible tags.

    It is used by other exceptions in the typechecked package to provide
    standardized error tagging for easier identification and handling of specific error conditions.
    and is used to create exceptions with specific tags for error handling and identification.
    with this base class.

    Example:

        from testspec import TaggedException

        class MyTaggedValueError(TaggedException[ValueError]):
        '''A tagged exception that is a specialized ValueError.'''

        raise MyTaggedValueError("An error occurred", tag=MyErrorTags.SOME_ERROR)

    :param args: Positional arguments to pass to the base exception's constructor.
    :type args: Any
    :param tag: An Enum member representing the error code.
    :type tag: Enum
    :param kwargs: Keyword arguments to pass to the base exception's constructor.
    :type kwargs: Any
    :ivar tag_code: The tag code for the exception.
    :vartype tag_code: Enum
    """
    __test__ = False  # Prevent pytest from trying to collect this class as a test case

    def __init__(self, *args: Any, tag: Enum, **kwargs: Any) -> None:
        """
        Initializes the exception with a mandatory tag.

        :param args: Positional arguments to pass to the base exception's constructor.
        :type args: Any
        :param tag: An Enum member representing the error code.
        :type tag: Enum
        :param kwargs: Keyword arguments to pass to the base exception's constructor.
        :type kwargs: Any
        """
        if not isinstance(tag, Enum):
            raise TypeError("Missing or wrong type 'tag' argument (must be Enum)")
        self.tag_code = tag
        super().__init__(*args, **kwargs)



from collections.abc import Sequence
from pathlib import Path, PosixPath, WindowsPath  # noqa: F401  # Needed for repr eval

import pytest

from autopypath._config_py_path._config._toml import _TomlConfig
from autopypath._exceptions import AutopypathError
from autopypath._marker_type import _MarkerType


def test_toml_config_init(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    marker_file = repo_root / 'filename.txt'
    marker_file.touch()
    marker_dir = repo_root / 'dirname'
    marker_dir.mkdir()
    (repo_root / 'src').mkdir()  # One of the two default paths so we can test both with and without presence
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
load_strategy = 'prepend'
path_resolution_order = ['manual', 'autopypath', 'pyproject']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except Exception as e:
        pytest.fail(f'TOML_001 Initialization of TomlConfig failed with exception: {e}')


def test_toml_missing_section(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.other_tool]
paths = ['src', 'lib']
""")

    config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    assert config.load_strategy is None, 'TOML_002 Expected load_strategy to be None when section is missing'
    assert config.paths is None, 'TOML_003 Expected paths to be None when section is missing'
    assert config.path_resolution_order is None, (
        'TOML_004 Expected path_resolution_order to be None when section is missing'
    )
    assert config.repo_markers is None, 'TOML_005 Expected marker_files to be None when section is missing'


def test_toml_missing_file(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()

    toml_filename = 'non_existent.toml'
    config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    assert config.no_file_found, 'TOML_MISSING_001 Expected no_file_found to be True when TOML file does not exist'


def test_toml_invalid_repo_marker_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = 'should_be_a_table_not_a_string'
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_007 Expected AutopypathError when repo_markers has invalid syntax')


def test_toml_invalid_path_resolution_order_value(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
path_resolution_order = ['invalid_value']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_008 Expected AutopypathError when path_resolution_order has an invalid value')


def test_toml_invalid_load_strategy_value(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
load_strategy = 'invalid_strategy'
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_009 Expected AutopypathError when load_strategy has an invalid value')


def test_toml_invalid_load_strategy_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
load_strategy = ['should_be_a_string_not_a_list']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_010 Expected AutopypathError when load_strategy has invalid syntax')


def test_toml_invalid_paths_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = 'should_be_a_list_not_a_string'
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_011 Expected AutopypathError when paths has invalid syntax')


def test_toml_invalid_repo_markers_value(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = {'.git': 'should_be_MarkerType_not_general_string'}
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_012 Expected AutopypathError when repo_markers has an invalid value')


def test_toml_invalid_paths_value(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 123]
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_013 Expected AutopypathError when paths has an invalid type of value')


def test_toml_invalid_path_resolution_order_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
path_resolution_order = 'should_be_a_list_not_a_string'
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_014 Expected AutopypathError when path_resolution_order has invalid syntax')


def test_toml_invalid_repo_markers_filename_type(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = {123='DIR'}
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_015 Expected AutopypathError when repo_markers has an invalid type for the filename')


def test_toml_invalid_repo_markers_filename_value(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = {'*invalid*': 'DIR'}
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_016 Expected AutopypathError when repo_markers has an invalid filename value')


def test_toml_misconfigured_section_name(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool..autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_017 Expected AutopypathError when toml_section has an invalid syntax')


def test_toml_incorrect_section_table_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""

[tool]

autopypath = 'hello'

""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_018 Expected AutopypathError when toml_section is misdefined in the TOML file')


def test_toml_repo_markers_table_syntax(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = ['.git', 'DIR', 'setup.py', 'FILE']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_019 Expected AutopypathError when repo_markers is not a table/dictionary')


def test_toml_repo_markers_table_value_type(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = {'.git'=['should_be_string_not_list']}
""")
    try:
        config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
        pytest.fail('TOML_020 Expected AutopypathError when repo_markers has '
                    f'non-string value type: {config.repo_markers}')
    except AutopypathError:
        pass


def test_toml_repr(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""[tool.autopypath]
paths = ['src', 'lib']
""")
    config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    repr_str = repr(config)
    assert 'TomlConfig' in repr_str, 'TOML_021 Expected __repr__ to contain class name'
    assert 'paths' in repr_str, 'TOML_022 Expected __repr__ to contain paths attribute'
    assert 'repo_markers' in repr_str, 'TOML_023 Expected __repr__ to contain repo_markers attribute'
    assert 'load_strategy' in repr_str, 'TOML_024 Expected __repr__ to contain load_strategy attribute'
    assert 'path_resolution_order' in repr_str, 'TOML_025 Expected __repr__ to contain path_resolution_order attribute'
    try:
        round_trip_config = eval(repr_str)
    except Exception as e:
        pytest.fail(f'TOML_026 Expected __repr__ to be evaluable without exception, but got: {e}\nRepr: {repr_str}')
    assert round_trip_config == config, (
        f'TOML_027 Expected __repr__ to be evaluable to the same object value: {repr_str}'
    )


def test_toml_str(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""[tool.autopypath]
paths = ['src', 'lib']
""")
    config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    str_repr = str(config)
    assert 'TomlConfig' in str_repr, 'TOML_028 Expected __str__ to contain class name'
    assert 'paths' in str_repr, 'TOML_029 Expected __str__ to contain paths attribute'
    assert 'repo_markers' in str_repr, 'TOML_030 Expected __str__ to contain repo_markers attribute'
    assert 'load_strategy' in str_repr, 'TOML_031 Expected __str__ to contain load_strategy attribute'
    assert 'path_resolution_order' in str_repr, 'TOML_032 Expected __str__ to contain path_resolution_order attribute'


def test_toml_good_repo_markers(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
repo_markers = {'.git'='dir', 'setup.py'='file'}
""")
    try:
        config = _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except Exception as e:
        pytest.fail(f'TOML_033 Initialization of TomlConfig with valid repo_markers failed with exception: {e}')
    assert config.repo_markers == {'.git': _MarkerType.DIR, 'setup.py': _MarkerType.FILE}, (
        f'TOML_034 Expected repo_markers to be correctly parsed, got: {config.repo_markers}'
    )


def test_toml_filename_value(tmp_path: Path) -> None:
    """Test that providing an invalid toml_filename raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml '  # Exists, but invalid name due to trailing space
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")

    toml_section = 'tool.autopypath'
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section=toml_section)
    except AutopypathError:
        return
    pytest.fail('TOML_035 Expected AutopypathError when toml_filename has invalid value')


def test_toml_filename_type(tmp_path: Path) -> None:
    """Test that providing a non-string toml_filename raises AutopypathError."""
    toml_section = 'tool.autopypath'

    # Create a temporary TOML file to satisfy the existence check
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")

    bad_toml_filename = 123  # Invalid type

    try:
        _TomlConfig(
            repo_root_path=repo_root,
            toml_filename=bad_toml_filename,  # type: ignore
            toml_section=toml_section,
        )
    except AutopypathError:
        return
    pytest.fail('TOML_036 Expected AutopypathError when toml_filename has invalid type')


def test_tome_filename_not_toml_suffix(tmp_path: Path) -> None:
    """Test that providing a toml_filename without .toml suffix raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'config.txt'  # Invalid suffix
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")
    toml_section = 'tool.autopypath'

    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section=toml_section)
    except AutopypathError:
        return
    pytest.fail('TOML_037 Expected AutopypathError when toml_filename does not have .toml suffix')


def test_toml_section_cannot_be_empty_string(tmp_path: Path) -> None:
    """Test that providing an empty string for toml_section raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")
    bad_toml_section = ''  # Invalid empty string for section name
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section=bad_toml_section)
    except AutopypathError:
        return
    pytest.fail('TOML_038 Expected AutopypathError when toml_section is an empty string')


def test_toml_section_type(tmp_path: Path) -> None:
    """Test that providing a non-string toml_section raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")
    bad_toml_section = 456  # Invalid type for section name
    try:
        _TomlConfig(
            repo_root_path=repo_root,
            toml_filename=toml_filename,
            toml_section=bad_toml_section,  # type: ignore
        )
    except AutopypathError:
        return
    pytest.fail('TOML_039 Expected AutopypathError when toml_section has invalid type')


def test_toml_section_invalid_adjacent_dots(tmp_path: Path) -> None:
    """Test that providing a toml_section with adjacent special chars raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""
[tool.autopypath]
paths = ['src', 'lib']
""")
    bad_passed_sections: list[str] = []
    for bad_toml_section in [
        'tool..autopypath',
        'tool.-autopypath',
        'tool._autopypath',
        'tool--autopypath',
        'tool-.autopypath',
        'tool-_autopypath',
        'tool__autopypath',
        'tool_.autopypath',
        'tool_-autopypath',
    ]:
        try:
            _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section=bad_toml_section)
        except AutopypathError:
            continue
        bad_passed_sections.append(bad_toml_section)
    if bad_passed_sections:
        bad_items = f'{bad_passed_sections!r}'
        pytest.fail(
            'TOML_040 Expected AutopypathError when toml_section is invalid: '
            f'{bad_items} unexpectedly passed validation.'
        )


def test_toml_section_invalid_start_end_chars(tmp_path: Path) -> None:
    """Test that providing a toml_section with invalid start/end chars raises AutopypathError."""
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""[tool.autopypath]
paths = ['src', 'lib']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='-tool.autopypath')
    except AutopypathError:
        return

    pytest.fail('TOML_041 Expected AutopypathError when toml_section starts with invalid character')


def test_toml_repo_root_path_is_none() -> None:
    """Test that providing None for repo_root_path creates special behavior."""
    toml_filename = 'toml_file.toml'
    toml_section = 'tool.autopypath'

    config = _TomlConfig(repo_root_path=None, toml_filename=toml_filename, toml_section=toml_section)

    assert config.paths is None, 'TOML_042 Expected paths to be None when repo_root_path is None'
    assert config.repo_markers is None, 'TOML_043 Expected repo_markers to be None when repo_root_path is None'
    assert config.load_strategy is None, 'TOML_044 Expected load_strategy to be None when repo_root_path is None'
    assert config.path_resolution_order is None, (
        'TOML_045 Expected path_resolution_order to be None when repo_root_path is None'
    )
    assert str(config.toml_filepath) == '<_NoPath>', (
        'TOML_046 Expected toml_filepath to be <_NoPath> when repo_root_path is None'
    )
    assert config.toml_section == toml_section, (
        'TOML_047 Expected toml_section to be set correctly even when repo_root_path is None'
    )


def test_toml_config_paths(tmp_path: Path) -> None:
    # Create toml file with autopypath configuration
    src_path = tmp_path / 'src'
    tests_path = tmp_path / 'tests'
    toml_path = tmp_path / 'config.toml'
    toml_path.write_text("""
[tool.autopypath]
paths = ["src", "tests"]
""")
    config = _TomlConfig(repo_root_path=tmp_path, toml_filename='config.toml', toml_section='tool.autopypath')
    assert isinstance(config.paths, Sequence), 'TOML_048 Expected paths to be a sequence type'
    assert len(config.paths) == 2, 'TOML_049 Expected paths list to have length 2'
    assert str(config.paths[0].name) == 'src', 'TOML_050 Expected first path to be "src"'
    assert str(config.paths[1].name) == 'tests', 'TOML_051 Expected second path to be "tests"'
    assert config.paths[0] == src_path, 'TOML_052 Expected first path to be match the path to src directory'
    assert config.paths[1] == tests_path, 'TOML_053 Expected second path to be match the path to tests directory'

def test_toml_path_cannot_contain_backslashes(tmp_path: Path) -> None:
    repo_root = tmp_path / 'my_repo'
    repo_root.mkdir()
    toml_filename = 'toml_file.toml'
    toml_path = repo_root / toml_filename
    toml_path.write_text("""[tool.autopypath]
paths = ['src\\lib']
""")
    try:
        _TomlConfig(repo_root_path=repo_root, toml_filename=toml_filename, toml_section='tool.autopypath')
    except AutopypathError:
        return
    pytest.fail('TOML_054 Expected AutopypathError when path contains backslashes')



"""Tests for :mod:`autopypath` module."""
import importlib
import logging
import sys
from pathlib import Path

import pytest


def test_autopypath_import_log_level(tmp_path: Path) -> None:
    """Test that importing autopypath sets log level to INFO."""
    repo_path = tmp_path / 'repo'
    repo_path.mkdir()
    autopypath_toml_path = repo_path / 'autopypath.toml'
    autopypath_toml_path.write_text("""
[tool.autopypath]
paths = ["src", "tests"]
load_strategy = "prepend"
path_resolution_order = ["manual", "autopypath"]
repo_markers = {".git" = "dir", "autopypath.toml" = "file"}
""")

    sys.modules.pop('autopypath', None)
    import autopypath
    from autopypath._log import _log

    assert _log.level == logging.NOTSET, 'AUTOPYPATH_001 Log level should be set to NOTSET upon importing autopypath'
    assert autopypath._context_name != '__main__', (
        'AUTOPYPATH_002 autopypath._context_name should not be __main__ when imported from inside a function; '
        f'got context name: {autopypath._context_name!r}'
    )
    assert autopypath._path_adjusted is False, (
        'AUTOPYPATH_003 autopypath._path_adjusted should be False when autopypath is '
        'imported from non-__main__ context.'
    )


def test_forced_main_execution(tmp_path: Path) -> None:
    """Test autopypath behavior when run as __main__."""

    # "Do not cite the Deep Magic to me, Witch! I was there when it was written."
    #
    # The following code forcibly sets __name__ to '__main__' to simulate running as a script.
    # This is deep magic and should not be used outside of testing and might break in future Python versions.
    # It has been tested on Python 3.9 through 3.14, and PyPy 3.10 and 3.11 as of Jan. 2026

    repo_path = tmp_path / 'repo'
    repo_path.mkdir()
    autopypath_toml_path = repo_path / 'autopypath.toml'
    autopypath_toml_path.write_text("""
[tool.autopypath]
paths = ["src", "tests"]
load_strategy = "prepend"
path_resolution_order = ["manual", "autopypath"]
repo_markers = {".git" = "dir", "autopypath.toml" = "file"}
""")

    # We do some shenanigans here to forcibly reload autopypath with __name__ set to '__main__'.

    # "Do not cite the Deep Magic to me, Witch! I was there when it was written."
    #
    # The following code forcibly sets __name__ to '__main__' to simulate running as a script.
    # This is deep magic and should not be used outside of testing and may break in future Python versions.
    # It has been tested on Python 3.9 through 3.14, and PyPy 3.10 and 3.11 as of Jan. 2026
    global __name__
    original_name = __name__
    try:
        __name__ = '__main__'
        sys.modules.pop('autopypath', None)
        import autopypath
        from autopypath._log import _log
    finally:
        __name__ = original_name

    assert _log.level == logging.NOTSET, 'AUTOPYPATH_001 Log level should be set to NOTSET upon importing autopypath'
    assert autopypath._context_name == '__main__', (
        'AUTOPYPATH_002 autopypath._context_name should be __main__ because we invoked deep magic; '
        f'got context name: {autopypath._context_name!r}'
    )
    assert autopypath._path_adjusted is True, (
        'AUTOPYPATH_003 autopypath._path_adjusted should be True when autopypath is '
        'imported from our forced __main__ context.'
    )


def test_import_raises_when_context_unknown(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that importing autopypath raises AutopypathError when context frame is unknown."""
    global __name__
    original_name = __name__
    try:
        __name__ = '__main__'
        sys.modules.pop('autopypath', None)
        import autopypath

        # Patch the frame introspection to simulate inability to find context frame
        monkeypatch.setattr("inspect.currentframe", lambda: None)

        with pytest.raises(autopypath.AutopypathError, match="could not determine context file"):
            importlib.reload(autopypath)
    finally:
        __name__ = original_name

# For direct manual execution of tests as a script.
#
# This is what pytest would do to set up sys.path for testing when run from the command line,
# and what autopypath is supposed to help with by removing the need for hardcoding manually
# determined sys.path tweaks that have to be made on a script-by-script basis.
#
# Note that when run via pytest, __name__ will NOT be '__main__', so autopypath
# will not adjust sys.path automatically during the test runs.
#
# When using autopypath in real scripts, it adjusts sys.path automatically
# when the script is run directly (i.e., when __name__ == '__main__')
# without needing any special test harness code to set up sys.path.
#
# It becomes
#
#   import autopypath
#   import pytest
#
#   # rest of your script code here
#
#   if __name__ == '__main__':
#       pytest.main(__file__)
#
# And it **just works** when run directly at the command line as `python your_script.py`
#


if __name__ == '__main__':
    repo_dir = Path(__file__).parent.parent
    sys.path = [str(repo_dir / 'src'), str(repo_dir / 'tests')] + sys.path

    pytest.main([__file__])




"""Test context detection functionality."""

import inspect
from pathlib import Path
from types import FrameType
from typing import Union

import pytest

from autopypath._context import _context_frameinfo


def test_context_frameinfo_basic() -> None:
    """Test that the context frame info is correctly identified."""
    context_info = _context_frameinfo()
    assert context_info is not None, 'CONTEXT_001 Context frame info should not be None'
    context_path, context_name = context_info
    assert isinstance(context_path, type(Path())), 'CONTEXT_002 Context path should be a Path object'
    assert isinstance(context_name, str), 'CONTEXT_003 Context name should be a string'
    # The test file itself should be the context
    assert context_name == __name__, 'CONTEXT_004 Context name should match the test module name'
    assert context_path.name == 'test_context.py', 'CONTEXT_005 Context path should point to this test file'


def test_context_frameinfo_no_autopypath() -> None:
    """Test that the context frame info is identified when autopypath is not in the stack."""

    def dummy_function() -> Union[tuple[Path, str], None]:
        return _context_frameinfo()

    context_info = dummy_function()
    assert context_info is not None, 'CONTEXT_006 Context frame info should not be None'
    context_path, context_name = context_info
    assert context_name == __name__, 'CONTEXT_007 Context name should match the test module name'
    assert context_path.name == 'test_context.py', 'CONTEXT_008 Context path should point to this test file'


def test_context_frameinfo_currentframe_none(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that _context_frameinfo returns None if inspect.currentframe() is None."""
    from autopypath._context import _context_frameinfo

    monkeypatch.setattr(inspect, 'currentframe', lambda: None)
    result = _context_frameinfo()
    assert result is None, 'CONTEXT_009 Should return None if currentframe() is None'


def test_context_frameinfo_no_file_in_globals(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that _context_frameinfo returns None if __file__ is missing."""
    class DummyCode:
        co_name = "dummy"

    class DummyFrame:
        f_globals = {"__name__": "dummy"}  # no __file__
        f_code = DummyCode()
        f_back: Union[FrameType, None] = None

    monkeypatch.setattr(inspect, "currentframe", lambda: DummyFrame())
    result = _context_frameinfo()
    assert result is None, "CONTEXT_010 Should return None if __file__ is missing in globals"



"""These are special internal types used by autopypath."""
# mypy: disable-error-code=override
# pyright: reportIncompatibleMethodOverride=false

from pathlib import Path
from typing import TYPE_CHECKING, final

from .._typing import Never, TypeAlias


class _NotSupported(NotImplementedError):
    """Exception raised when an unsupported operation is attempted on _NoPath.

    This exception indicates that the attempted operation is not supported.
    """


if TYPE_CHECKING:
    # During static analysis, make the "disabled" parameter un-callable:
    # nothing can ever satisfy a Never-typed argument.
    _UsagePreventedType: TypeAlias = Never
else:

    @final
    class _UsagePreventedType:
        """A type to indicate that a type is deliberately broken to prevent use of a method.

        This type is used as a type hint for method parameters in the :class:`_NoPath` class
        to indicate that these methods should not be used and will raise a :class:`_NotSupported`
        exception if called.

        It deliberately breaks the method signature to prevent accidental usage.

        This helps type checkers identify incorrect usage of these methods before runtime.
        """


_PathType: type[Path] = type(Path())
"""A type alias for the pathlib.Path type.

Generally, you cannot directly inherit from pathlib.Path because it is a C extension type.
This alias allows us to work around that limitation.
"""


_NOT_SUPPORTED_ERR: str = '_NoPath does not support this operation.'


@final
class _NoPath(_PathType):  # type: ignore[valid-type,misc]  # Magic to inherit from pathlib.Path
    """A sentinel Path type representing the absence of a path.

    This class is used to indicate that no valid path is available or applicable
    while still conforming to the expected Path type in type hints and function
    signatures without having to juggle ``None | Path`` types in those signatures.

    It inherits from :class:`~pathlib.Path` to maintain compatibility with path operations
    and type checking, but it signifies a 'no path' state and cannot be used to perform
    file system and other path-related operations.
    """

    def __hash__(self) -> int:
        """Hash for _NoPath objects always returns 0."""
        return 0

    def __repr__(self) -> str:
        """Representation of _NoPath."""
        return '<_NoPath>'

    def __str__(self) -> str:
        """String representation of _NoPath."""
        return '<_NoPath>'

    def __eq__(self, other: object) -> bool:
        """Equality comparison for _NoPath objects.

        :param object other: Another object to compare with.
        :return bool: True if both are _NoPath instances, False otherwise.
        """
        return isinstance(other, _NoPath)

    # methods overridden to prevent filesystem access and other path-related operations
    # all raise _NotSupported when called. This is maintained for type compatibility only.
    # and as a guardrail against misuse of this sentinel type for actual path operations.
    # The primary purpose of this class is to act as a sentinel value, not
    # to perform path operations and thus path operations are disabled.

    def __fspath__(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def __truediv__(self, key: object) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def __rtruediv__(self, key: object) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def exists(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_file(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_dir(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def open(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def read_text(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def read_bytes(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def write_text(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def write_bytes(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def mkdir(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def rmdir(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def unlink(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def rename(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def replace(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def touch(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def stat(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def chmod(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def lstat(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def owner(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def group(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def readlink(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def symlink_to(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def hardlink_to(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def absolute(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def resolve(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def samefile(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def expanduser(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def with_name(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def with_suffix(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def relative_to(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_absolute(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_reserved(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def joinpath(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def match(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def parent(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def parents(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def parts(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def drive(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def root(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def anchor(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def name(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def suffix(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def suffixes(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @property
    def stem(self) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def as_posix(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def as_uri(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_mount(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_symlink(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_block_device(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_char_device(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_fifo(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def is_socket(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def iterdir(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def glob(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    def rglob(self, _disabled: _UsagePreventedType) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @classmethod
    def cwd(cls) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)

    @classmethod
    def home(cls) -> Never:
        raise _NotSupported(_NOT_SUPPORTED_ERR)



"""Helpers for running doctest benchmarks and validating their output."""
import difflib
import os
import re
import subprocess
import sys
from typing import List

# --- Path Configuration ---
# The 'documentation' directory is our consistent root for all paths.
HELPER_DIR = os.path.dirname(os.path.abspath(__file__))
DOCS_ROOT = os.path.dirname(HELPER_DIR)
# --------------------------


def _sanitize_output(raw_text: str) -> str:
    """Internal helper to sanitize benchmark output text."""
    sanitized_lines = []
    for line in raw_text.splitlines():
        if line.strip().startswith('│'):
            line = re.sub(r'\b\d[\d.,]*%?\b', '...', line)
        sanitized_lines.append(line)
    return "\n".join(sanitized_lines)


def run_script_and_get_raw_output(script_path: str, args: List[str]) -> str:
    """
    Executes a script and returns its raw, unsanitized stdout.
    The script path is relative to the 'documentation' directory.
    """
    command = [sys.executable, script_path] + args
    env = os.environ.copy()
    env['COLUMNS'] = '500'

    # The subprocess needs the project root in its PYTHONPATH to find 'simplebench'.
    project_root = os.path.dirname(DOCS_ROOT)
    env['PYTHONPATH'] = f"{project_root}{os.pathsep}{env.get('PYTHONPATH', '')}"

    # Run the subprocess from the documentation root for consistent pathing.
    result = subprocess.run(
        command, capture_output=True, text=True, check=False,
        cwd=DOCS_ROOT, env=env
    )
    if result.stderr:
        return f"--- STDERR ---\n{result.stderr}"
    return result.stdout


def assert_benchmark_output(actual_raw: str, expected_output_file: str):
    """
    Sanitizes and compares benchmark output against an expected output file.
    The file path is relative to the 'documentation' directory.
    """
    # Construct the full path to the expected output file from the docs root.
    full_expected_path = os.path.join(DOCS_ROOT, expected_output_file)

    with open(full_expected_path, encoding="utf-8") as fh:
        expected_raw = fh.read()

    actual_sanitized = _sanitize_output(actual_raw)
    expected_sanitized = _sanitize_output(expected_raw)

    # To compare, normalize all whitespace in both strings.
    actual_normalized = " ".join(actual_sanitized.split())
    expected_normalized = " ".join(expected_sanitized.split())

    if actual_normalized != expected_normalized:
        diff = difflib.unified_diff(
            expected_sanitized.splitlines(keepends=True),
            actual_sanitized.splitlines(keepends=True),
            fromfile=f'expected (from {expected_output_file})',
            tofile='actual (live output)',
        )
        error_message = "Sanitized output does not match expected output (ignoring whitespace):\n" + "".join(diff)
        raise AssertionError(error_message)


if __name__ == "__main__":
    # This block allows the script to be used as a command-line tool to
    # generate the "golden master" output files for doctests.
    # Run this script from the 'documentation' directory.
    if len(sys.argv) < 3:
        print(f"Usage: python {sys.argv[0]} <output_file> <script_to_run> [args...]")
        print("  (Run from 'documentation/' directory, paths are relative to it)")
        sys.exit(1)

    output_file_path = sys.argv[1]
    script_to_run = sys.argv[2]
    script_args = sys.argv[3:]

    print(f"Generating golden master for '{script_to_run}'...")

    raw_output = run_script_and_get_raw_output(script_to_run, script_args)

    with open(output_file_path, "w", encoding="utf-8") as f:
        f.write(raw_output)

    print(f"Successfully generated golden master file: {output_file_path}")



"""Custom autopypath configurations.

**Custom Configuration Interface**
----------------------------------

This module provides a function :func:`configure_pypath` that allows users to
customize how the :data:`sys.path` is set up according to their specific needs.

It provides detailed control over repository markers, additional paths, load strategy,
resolution order, logging level, and strictness of configuration.

By importing the :mod:`autopypath.custom` submodule instead of :mod:`autopypath`,
no automatic adjustments to :data:`sys.path` are made. Instead, users can call
:func:`configure_pypath` with their desired parameters to set up the PYTHONPATH
according to their requirements.

It still must be imported from the `__main__` context to function. The call
to :func:`configure_pypath` will be a no-op if imported from a non-`__main__` context.

This prevents unintended side effects in modules that are not the main script such
as when running unit tests or interactive sessions.

The call to :func:`configure_pypath` must be place early in the execution of the script,
before any other imports that depend on the adjusted :data:`sys.path`.

**Example Usage**
-----------------

.. code-block:: python

    import logging
    from autopypath.custom import configure_pypath

    configure_pypath(
        repo_markers={'pyproject.toml': 'file', '.git': 'dir'},
        paths=['src', 'lib'],
        load_strategy='prepend',
        path_resolution_order=['autopypath', 'pyproject'],
        log_level=logging.INFO,
        strict=True,
    )

    import mymodule  # This import can now rely on the adjusted sys.path
"""

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Optional, Union

from .._config_py_path import _ConfigPyPath
from .._context import _context_frameinfo
from .._exceptions import AutopypathError
from .._log import _log
from .._types import LoadStrategyLiterals, PathResolutionLiterals, RepoMarkerLiterals

__all__ = ['configure_pypath', 'AutopypathError']

_NOT_MAIN_CONTEXT_WARNING = 'autopypath.custom imported from non-__main__ context; no sys.path changes will be applied.'
_context_file: Optional[Path] = None
"""This is the file path of the script that imported this module, if available."""
_context_name: Optional[str] = None
"""This is the __name__ of the script that imported this module, if available."""
_context_info: Optional[tuple[Path, str]] = _context_frameinfo()
"""Context information (path, __name__) tuple about the importing script, if available."""

if _context_info is not None:
    _context_file, _context_name = _context_info
    if _context_name != '__main__':
        message = _NOT_MAIN_CONTEXT_WARNING.format(_context_name)
        _log.debug(message)

else: # unable to determine context info
    _context_file = None
    _context_name = None
    _log.error('could not determine context file; no sys.path changes will be applied.')
    raise AutopypathError('could not determine context file; no sys.path changes will be applied.')

_ran_once: bool = False
"""Indicates whether :func:`configure_pypath` has been called already."""

def configure_pypath(
    *,
    repo_markers: Optional[Mapping[str, RepoMarkerLiterals]] = None,
    paths: Optional[Sequence[Union[Path, str]]] = None,
    load_strategy: Optional[LoadStrategyLiterals] = None,
    path_resolution_order: Optional[Sequence[PathResolutionLiterals]] = None,
    log_level: Optional[int] = None,
    strict: bool = False,
) -> None:
    """Configures the PYTHONPATH according to the provided parameters.

    Configures the :data:`sys.path` according to the provided parameters.

    This function allows customization of how the :data:`sys.path` is set up,
    including repository markers, additional paths, load strategy, and resolution order.

    :param Mapping[str, Literal['dir', 'file']] | None repo_markers: A mapping of file or directory names to
        their MarkerType used to identify the repository root. They can only be of type 'dir' or 'file'
        and must be names only (no paths). If None, the default repo markers are used.
    :param Sequence[Path | str] | None paths: A sequence of paths to include in the :data:`sys.path`.
        If passed as strings, the must be formatted as POSIX-style paths (e.g., 'src/utils') and
        cannot be absolute paths.
        If passed as :class:`pathlib.Path` objects, they can be either absolute or relative paths.
    :param Literal['prepend', 'prepend_highest_priority', 'replace'] | None load_strategy: The strategy
        for loading :data:`sys.path` entries.
    :param Sequence[Literal['manual', 'autopypath', 'pyproject']] | None path_resolution_order: The order
        in which to resolve :data:`sys.path` sources.
    :param log_level: Optional[int] = None
        The logging level to use during configuration. If None, the current log level is used.
    :param bool strict: If True, raises an error for conditions that would normally only log a warning.
        Default is False.

        Conditions that normally trigger logged warnings include:
            - Imported from a non-`__main__` context.

    :raises AutopypathError: If the context file cannot be determined or if `strict` is set to ``True``
        and a condition that would normally log a warning occurs.
    """
    global _ran_once
    if _ran_once:
        _log.info('configure_pypath has already been called once; subsequent calls are no-ops.')
        return

    if isinstance(log_level, int):  # Set as early as possible
        _log.setLevel(log_level)

    if _context_file is None:
        _log.error('could not determine context file; cannot configure sys.path.')
        raise AutopypathError('could not determine context file; cannot configure sys.path.')
    elif _context_name != '__main__':
        if strict:
            _log.error(_NOT_MAIN_CONTEXT_WARNING)
            raise AutopypathError(
                f'autopypath.custom imported from non-__main__ context ({_context_name}); cannot configure sys.path.'
            )
        _log.warning(_NOT_MAIN_CONTEXT_WARNING)
    else:
        _ConfigPyPath(
            context_file=_context_file,
            repo_markers=repo_markers,
            paths=paths,
            load_strategy=load_strategy,
            path_resolution_order=path_resolution_order,
            log_level=log_level,
            strict=strict,
        )
        _ran_once = True
        _log.debug('sys.path adjusted automatically for %s', _context_file)



"""Metadata for autopypath package.

This file contains shared metadata for the project.
It is parsed by the build system and should only contain simple assignments
(str, int, float, bool, None).
"""

__author__ = 'Jerilyn Franz'
__copyright__ = '2026, Jerilyn Franz'
__version__ = '0.9.1'
__release__ = '0.9.1'
__project__ = 'autopypath'
__license__ = 'Apache-2.0'
__url__ = 'https://github.com/snowhare/python-autopypath'
